"""

video_utils.py

Utilities for splitting, rendering, and assembling videos.

"""

#%% Constants, imports, environment

import os
import cv2
import glob
import json

from collections import defaultdict
from multiprocessing.pool import ThreadPool
from multiprocessing.pool import Pool
from tqdm import tqdm
from functools import partial

from md_utils import path_utils    
from md_visualization import visualization_utils as vis_utils

default_fourcc = 'h264'


#%% Path utilities

VIDEO_EXTENSIONS = ('.mp4','.avi','.mpeg','.mpg')

def is_video_file(s,video_extensions=VIDEO_EXTENSIONS):
    """
    Checks a file's extension against a set of known video file
    extensions to determine whether it's a video file.  Performs a
    case-insensitive comparison.
    
    Args:
        s (str): filename to check for probable video-ness
        video_extensions (list, optional): list of video file extensions
    
    Returns:
        bool: True if this looks like a video file, else False
    """
    
    ext = os.path.splitext(s)[1]
    return ext.lower() in video_extensions


def find_video_strings(strings):
    """
    Given a list of strings that are potentially video file names, looks for
    strings that actually look like video file names (based on extension).
    
    Args:
        strings (list): list of strings to check for video-ness
    
    Returns:
        list: a subset of [strings] that looks like they are video filenames
    """
    
    return [s for s in strings if is_video_file(s.lower())]


def find_videos(dirname, 
                recursive=False,
                convert_slashes=True,
                return_relative_paths=False):
    """
    Finds all files in a directory that look like video file names.
    
    Args:
        dirname (str): folder to search for video files
        recursive (bool, optional): whether to search [dirname] recursively
        convert_slashes (bool, optional): forces forward slashes in the returned files,
            otherwise uses the native path separator
        return_relative_paths (bool, optional): forces the returned filenames to be 
            relative to [dirname], otherwise returns absolute paths
    
    Returns:
        A list of filenames within [dirname] that appear to be videos
    """
    
    if recursive:
        files = glob.glob(os.path.join(dirname, '**', '*.*'), recursive=True)
    else:
        files = glob.glob(os.path.join(dirname, '*.*'))
        
    if return_relative_paths:
        files = [os.path.relpath(fn,dirname) for fn in files]

    if convert_slashes:
        files = [fn.replace('\\', '/') for fn in files]
    
    return find_video_strings(files)


#%% Function for rendering frames to video and vice-versa

# http://tsaith.github.io/combine-images-into-a-video-with-python-3-and-opencv-3.html

def frames_to_video(images, Fs, output_file_name, codec_spec=default_fourcc):
    """
    Given a list of image files and a sample rate, concatenates those images into
    a video and writes to a new video file.
    
    Args:
        images (list): a list of frame file names to concatenate into a video
        Fs (float): the frame rate in fps
        output_file_name (str): the output video file, no checking is performed to make
            sure the extension is compatible with the codec
        codec_spec (str, optional):  codec to use for encoding; h264 is a sensible default 
            and generally works on Windows, but when this fails (which is around 50% of the time 
            on Linux), mp4v is a good second choice
    """
    
    if codec_spec is None:
        codec_spec = 'h264'
        
    if len(images) == 0:
        return

    # Determine the width and height from the first image
    frame = cv2.imread(images[0])
    cv2.imshow('video',frame)
    height, width, channels = frame.shape

    # Define the codec and create VideoWriter object
    fourcc = cv2.VideoWriter_fourcc(*codec_spec)
    out = cv2.VideoWriter(output_file_name, fourcc, Fs, (width, height))

    for image in images:
        frame = cv2.imread(image)
        out.write(frame)

    out.release()
    cv2.destroyAllWindows()


def get_video_fs(input_video_file):
    """
    Retrieves the frame rate of [input_video_file].
    
    Args:
        input_video_file (str): video file for which we want the frame rate
        
    Returns:
        float: the frame rate of [input_video_file]
    """
    
    assert os.path.isfile(input_video_file), 'File {} not found'.format(input_video_file)    
    vidcap = cv2.VideoCapture(input_video_file)
    Fs = vidcap.get(cv2.CAP_PROP_FPS)
    vidcap.release()
    return Fs


def _frame_number_to_filename(frame_number):
    """
    Ensures that frame images are given consistent filenames.
    """
    
    return 'frame{:06d}.jpg'.format(frame_number)


def video_to_frames(input_video_file, output_folder, overwrite=True, 
                    every_n_frames=None, verbose=False):
    """
    Renders frames from [input_video_file] to a .jpg in [output_folder].
    
    With help from:
        
    https://stackoverflow.com/questions/33311153/python-extracting-and-saving-video-frames
    
    Args:
        input_video_file (str): video file to split into frames
        output_folder (str): folder to put frame images in
        overwrite (bool, optional): whether to overwrite existing frame images
        every_n_frames (int, optional): sample every Nth frame starting from the first frame;
            if this is None or 1, every frame is extracted
        verbose (bool, optional): enable additional debug console output
    
    Returns:
        tuple: length-2 tuple containing (list of frame filenames,frame rate)
    """
    
    assert os.path.isfile(input_video_file), 'File {} not found'.format(input_video_file)
    
    vidcap = cv2.VideoCapture(input_video_file)
    n_frames = int(vidcap.get(cv2.CAP_PROP_FRAME_COUNT))
    Fs = vidcap.get(cv2.CAP_PROP_FPS)
    
    # If we're not over-writing, check whether all frame images already exist
    if overwrite == False:
        
        missing_frame_number = None
        frame_filenames = []
        
        for frame_number in range(0,n_frames):
            
            if every_n_frames is not None:
                if frame_number % every_n_frames != 0:
                    continue
            
            frame_filename = _frame_number_to_filename(frame_number)
            frame_filename = os.path.join(output_folder,frame_filename)
            frame_filenames.append(frame_filename)
            if os.path.isfile(frame_filename):
                continue
            else:
                missing_frame_number = frame_number
                break
    
        # OpenCV seems to over-report the number of frames by 1 in some cases, or fails
        # to read the last frame; either way, I'm allowing one missing frame.
        allow_last_frame_missing = True
        
        if missing_frame_number is None or \
            (allow_last_frame_missing and (missing_frame_number == n_frames-1)):
            if verbose:
                print('Skipping video {}, all output frames exist'.format(input_video_file))
            return frame_filenames,Fs
        else:
            pass
            # print("Rendering video {}, couldn't find frame {}".format(
            #    input_video_file,missing_frame_number))
    
    # ...if we need to check whether to skip this video entirely
        
    if verbose:
        print('Reading {} frames at {} Hz from {}'.format(n_frames,Fs,input_video_file))

    frame_filenames = []

    # for frame_number in tqdm(range(0,n_frames)):
    for frame_number in range(0,n_frames):

        success,image = vidcap.read()
        if not success:
            assert image is None
            if verbose:
                print('Read terminating at frame {} of {}'.format(frame_number,n_frames))
            break

        if every_n_frames is not None:
            if frame_number % every_n_frames != 0:
                continue
            
        frame_filename = _frame_number_to_filename(frame_number)
        frame_filename = os.path.join(output_folder,frame_filename)
        frame_filenames.append(frame_filename)
        
        if overwrite == False and os.path.isfile(frame_filename):
            # print('Skipping frame {}'.format(frame_filename))
            pass            
        else:
            try:
                if frame_filename.isascii():
                    cv2.imwrite(os.path.normpath(frame_filename),image)
                else:
                    is_success, im_buf_arr = cv2.imencode('.jpg', image)
                    im_buf_arr.tofile(frame_filename)
                assert os.path.isfile(frame_filename), \
                    'Output frame {} unavailable'.format(frame_filename)
            except KeyboardInterrupt:
                vidcap.release()
                raise
            except Exception as e:
                print('Error on frame {} of {}: {}'.format(frame_number,n_frames,str(e)))

    if verbose:
        print('\nExtracted {} of {} frames'.format(len(frame_filenames),n_frames))

    vidcap.release()    
    return frame_filenames,Fs

# ...def video_to_frames(...)


def _video_to_frames_for_folder(relative_fn,input_folder,output_folder_base,every_n_frames,overwrite,verbose):
    """
    Internal function to call video_to_frames in the context of video_folder_to_frames; 
    makes sure the right output folder exists, then calls video_to_frames.
    """    
    
    input_fn_absolute = os.path.join(input_folder,relative_fn)
    assert os.path.isfile(input_fn_absolute),\
        'Could not find file {}'.format(input_fn_absolute)

    # Create the target output folder
    output_folder_video = os.path.join(output_folder_base,relative_fn)
    os.makedirs(output_folder_video,exist_ok=True)

    # Render frames
    # input_video_file = input_fn_absolute; output_folder = output_folder_video
    frame_filenames,fs = video_to_frames(input_fn_absolute,output_folder_video,
                                         overwrite=overwrite,every_n_frames=every_n_frames,
                                         verbose=verbose)
    
    return frame_filenames,fs


def video_folder_to_frames(input_folder, output_folder_base, 
                           recursive=True, overwrite=True,
                           n_threads=1, every_n_frames=None,
                           verbose=False, parallelization_uses_threads=True):
    """
    For every video file in input_folder, creates a folder within output_folder_base, and 
    renders frame of that video to images in that folder.
    
    Args:
        input_folder (str): folder to process
        output_folder_base (str): root folder for output images; subfolders will be
            created for each input video
        recursive (bool, optional): whether to recursively process videos in [input_folder]
        overwrite (bool, optional): whether to overwrite existing frame images
        n_threads (int, optional): number of concurrent workers to use; set to <= 1 to disable
            parallelism
        every_n_frames (int, optional): sample every Nth frame starting from the first frame;
            if this is None or 1, every frame is extracted
        verbose (bool, optional): enable additional debug console output
        parallelization_uses_threads (bool, optional): whether to use threads (True) or
            processes (False) for parallelization; ignored if n_threads <= 1
            
    Returns:
        tuple: a length-3 tuple containing:
            - list of lists of frame filenames; the Nth list of frame filenames corresponds to 
              the Nth video
            - list of video frame rates; the Nth value corresponds to the Nth video
            - list of video filenames    
    """
    
    # Recursively enumerate video files
    input_files_full_paths = find_videos(input_folder,recursive=recursive)
    print('Found {} videos in folder {}'.format(len(input_files_full_paths),input_folder))
    if len(input_files_full_paths) == 0:
        return [],[],[]
    
    input_files_relative_paths = [os.path.relpath(s,input_folder) for s in input_files_full_paths]
    input_files_relative_paths = [s.replace('\\','/') for s in input_files_relative_paths]
    
    os.makedirs(output_folder_base,exist_ok=True)    
    
    frame_filenames_by_video = []
    fs_by_video = []
    
    if n_threads == 1:
        # For each video
        #
        # input_fn_relative = input_files_relative_paths[0]
        for input_fn_relative in tqdm(input_files_relative_paths):
        
            frame_filenames,fs = \
                _video_to_frames_for_folder(input_fn_relative,input_folder,output_folder_base,
                                            every_n_frames,overwrite,verbose)
            frame_filenames_by_video.append(frame_filenames)
            fs_by_video.append(fs)
    else:
        if parallelization_uses_threads:
            print('Starting a worker pool with {} threads'.format(n_threads))
            pool = ThreadPool(n_threads)
        else:
            print('Starting a worker pool with {} processes'.format(n_threads))
            pool = Pool(n_threads)
        process_video_with_options = partial(_video_to_frames_for_folder, 
                                             input_folder=input_folder,
                                             output_folder_base=output_folder_base,
                                             every_n_frames=every_n_frames,
                                             overwrite=overwrite,
                                             verbose=verbose)
        results = list(tqdm(pool.imap(
            partial(process_video_with_options),input_files_relative_paths), 
                            total=len(input_files_relative_paths)))
        frame_filenames_by_video = [x[0] for x in results]
        fs_by_video = [x[1] for x in results]
        
    return frame_filenames_by_video,fs_by_video,input_files_full_paths
  
# ...def video_folder_to_frames(...)


class FrameToVideoOptions:
    """
    Options controlling the conversion of frame-level results to video-level results via
    frame_results_to_video_results()    
    """
    
    #: One-indexed indicator of which frame-level confidence value to use to determine detection confidence
    #: for the whole video, i.e. "1" means "use the confidence value from the highest-confidence frame"
    nth_highest_confidence = 1
    
    #: What to do if a file referred to in a .json results file appears not to be a 
    #: video; can be 'error' or 'skip_with_warning'
    non_video_behavior = 'error'
    
    
def frame_results_to_video_results(input_file,output_file,options=None):
    """
    Given an MD results file produced at the *frame* level, corresponding to a directory 
    created with video_folder_to_frames, maps those frame-level results back to the 
    video level for use in Timelapse.
    
    Preserves everything in the input .json file other than the images.
    
    Args:
        input_file (str): the frame-level MD results file to convert to video-level results
        output_file (str): the .json file to which we should write video-level results
        options (FrameToVideoOptions, optional): parameters for converting frame-level results
            to video-level results, see FrameToVideoOptions for details            
    """

    if options is None:
        options = FrameToVideoOptions()
        
    # Load results
    with open(input_file,'r') as f:
        input_data = json.load(f)

    images = input_data['images']
    detection_categories = input_data['detection_categories']
    
    ## Break into videos
    
    video_to_frames = defaultdict(list) 
    
    # im = images[0]
    for im in tqdm(images):
        
        fn = im['file']
        video_name = os.path.dirname(fn)
        if not is_video_file(video_name):
            if options.non_video_behavior == 'error':
                raise ValueError('{} is not a video file'.format(video_name))
            elif options.non_video_behavior == 'skip_with_warning':
                print('Warning: {} is not a video file'.format(video_name))
                continue
            else:
                raise ValueError('Unrecognized non-video handling behavior: {}'.format(
                    options.non_video_behavior))
        video_to_frames[video_name].append(im)
    
    print('Found {} unique videos in {} frame-level results'.format(
        len(video_to_frames),len(images)))
    
    output_images = []
    
    ## For each video...
    
    # video_name = list(video_to_frames.keys())[0]
    for video_name in tqdm(video_to_frames):
        
        frames = video_to_frames[video_name]
        
        all_detections_this_video = []
        
        # frame = frames[0]
        for frame in frames:
            if frame['detections'] is not None:
                all_detections_this_video.extend(frame['detections'])
            
        # At most one detection for each category for the whole video
        canonical_detections = []
            
        # category_id = list(detection_categories.keys())[0]
        for category_id in detection_categories:
            
            category_detections = [det for det in all_detections_this_video if \
                                   det['category'] == category_id]
            
            # Find the nth-highest-confidence video to choose a confidence value
            if len(category_detections) >= options.nth_highest_confidence:
                
                category_detections_by_confidence = sorted(category_detections, 
                                                           key = lambda i: i['conf'],reverse=True)
                canonical_detection = category_detections_by_confidence[options.nth_highest_confidence-1]
                canonical_detections.append(canonical_detection)
                                      
        # Prepare the output representation for this video
        im_out = {}
        im_out['file'] = video_name
        im_out['detections'] = canonical_detections
        
        # 'max_detection_conf' is no longer included in output files by default
        if False:
            im_out['max_detection_conf'] = 0
            if len(canonical_detections) > 0:
                confidences = [d['conf'] for d in canonical_detections]
                im_out['max_detection_conf'] = max(confidences)
        
        output_images.append(im_out)
        
    # ...for each video
    
    output_data = input_data
    output_data['images'] = output_images
    s = json.dumps(output_data,indent=1)
    
    # Write the output file
    with open(output_file,'w') as f:
        f.write(s)
    
# ...def frame_results_to_video_results(...)


#%% Test driver

if False:

    #%% Constants
    
    Fs = 30.01
    confidence_threshold = 0.75
    input_folder = 'z:\\'
    frame_folder_base = r'e:\video_test\frames'
    detected_frame_folder_base = r'e:\video_test\detected_frames'
    rendered_videos_folder_base = r'e:\video_test\rendered_videos'
    
    results_file = r'results.json'
    os.makedirs(detected_frame_folder_base,exist_ok=True)
    os.makedirs(rendered_videos_folder_base,exist_ok=True)
    
    
    #%% Split videos into frames
        
    frame_filenames_by_video,fs_by_video,video_filenames = \
        video_folder_to_frames(input_folder,frame_folder_base,recursive=True)
    
    
    #%% List image files, break into folders
    
    frame_files = path_utils.find_images(frame_folder_base,True)
    frame_files = [s.replace('\\','/') for s in frame_files]
    print('Enumerated {} total frames'.format(len(frame_files)))
    
    Fs = 30.01
    # Find unique folders
    folders = set()
    # fn = frame_files[0]
    for fn in frame_files:
        folders.add(os.path.dirname(fn))
    folders = [s.replace('\\','/') for s in folders]
    print('Found {} folders for {} files'.format(len(folders),len(frame_files)))
    
        
    #%% Load detector output
    
    with open(results_file,'r') as f:
        detection_results = json.load(f)
    detections = detection_results['images']
    detector_label_map = detection_results['detection_categories']
    for d in detections:
        d['file'] = d['file'].replace('\\','/').replace('video_frames/','')


    #%% Render detector frames
    
    # folder = list(folders)[0]
    for folder in folders:
        
        frame_files_this_folder = [fn for fn in frame_files if folder in fn]
        folder_relative = folder.replace((frame_folder_base + '/').replace('\\','/'),'')
        detection_results_this_folder = [d for d in detections if folder_relative in d['file']]
        print('Found {} detections in folder {}'.format(len(detection_results_this_folder),folder))
        assert len(frame_files_this_folder) == len(detection_results_this_folder)
        
        rendered_frame_output_folder = os.path.join(detected_frame_folder_base,folder_relative)
        os.makedirs(rendered_frame_output_folder,exist_ok=True)
        
        # d = detection_results_this_folder[0]
        for d in tqdm(detection_results_this_folder):
            
            input_file = os.path.join(frame_folder_base,d['file'])
            output_file = os.path.join(detected_frame_folder_base,d['file'])
            os.makedirs(os.path.dirname(output_file),exist_ok=True)
            vis_utils.draw_bounding_boxes_on_file(input_file,output_file,d['detections'],
                                                  confidence_threshold)
        
        # ...for each file in this folder
            
    # ...for each folder


    #%% Render output videos
            
    # folder = list(folders)[0]
    for folder in tqdm(folders):
        
        folder_relative = folder.replace((frame_folder_base + '/').replace('\\','/'),'')
        rendered_detector_output_folder = os.path.join(detected_frame_folder_base,folder_relative)
        assert os.path.isdir(rendered_detector_output_folder)
        
        frame_files_relative = os.listdir(rendered_detector_output_folder)
        frame_files_absolute = [os.path.join(rendered_detector_output_folder,s) \
                                for s in frame_files_relative]
        
        output_video_filename = os.path.join(rendered_videos_folder_base,folder_relative)
        os.makedirs(os.path.dirname(output_video_filename),exist_ok=True)
        
        original_video_filename = output_video_filename.replace(
            rendered_videos_folder_base,input_folder)
        assert os.path.isfile(original_video_filename)
        Fs = get_video_fs(original_video_filename)
                
        frames_to_video(frame_files_absolute, Fs, output_video_filename)

    # ...for each video



"""

add_locations_to_nacti.py

As of 10.2023, NACTI metadata only has very coarse location information (e.g. "Florida"),
but camera IDs are embedded in filenames.  This script pulls that information from filenames
and adds it to metadata.

"""

#%% Imports and constants

import os
import json
import shutil

from tqdm import tqdm
from collections import defaultdict

input_file = r'd:\lila\nacti\nacti_metadata.json.1.13\nacti_metadata.json'
output_file = r'g:\temp\nacti_metadata.1.14.json'


#%% Read metadata

with open(input_file,'r') as f:
    d = json.load(f)
    
assert d['info']['version'] == 1.13


#%% Map images to locations (according to the metadata)

file_name_to_original_location = {}

# im = dataset_labels['images'][0]
for im in tqdm(d['images']):
    file_name_to_original_location[im['file_name']] = im['location']

original_locations = set(file_name_to_original_location.values())

print('Found {} locations in the original metadata:'.format(len(original_locations)))
for loc in original_locations:
    print('[{}]'.format(loc))


#%% Map images to new locations

def path_to_location(relative_path):
    
    relative_path = relative_path.replace('\\','/')
    if relative_path in file_name_to_original_location:
        location_name = file_name_to_original_location[relative_path]
        if location_name == 'San Juan Mntns, Colorado':
            # "part0/sub000/2010_Unit150_Ivan097_img0003.jpg"
            tokens = relative_path.split('/')[-1].split('_')
            assert tokens[1].startswith('Unit')
            location_name = 'sanjuan_{}_{}_{}'.format(tokens[0],tokens[1],tokens[2])
        elif location_name == 'Lebec, California':
            # "part0/sub035/CA-03_08_13_2015_CA-03_0009738.jpg"
            tokens = relative_path.split('/')[-1].split('_')
            assert tokens[0].startswith('CA-') or tokens[0].startswith('TAG-')
            location_name = 'lebec_{}'.format(tokens[0])   
        elif location_name == 'Archbold, FL':
            # "part1/sub110/FL-01_01_25_2016_FL-01_0040421.jpg"
            tokens = relative_path.split('/')[-1].split('_')
            assert tokens[0].startswith('FL-')
            location_name = 'archbold_{}'.format(tokens[0])   
        else:
            assert location_name == ''
            tokens = relative_path.split('/')[-1].split('_')
            if tokens[0].startswith('CA-') or tokens[0].startswith('TAG-') or tokens[0].startswith('FL-'):
                location_name = '{}'.format(tokens[0])
            
    else:
        
        location_name = 'unknown'    
    
    # print('Returning location {} for file {}'.format(location_name,relative_path))          

    return location_name

file_name_to_updated_location = {}
updated_location_to_count = defaultdict(int)
for im in tqdm(d['images']):
    
    updated_location = path_to_location(im['file_name'])
    file_name_to_updated_location[im['file_name']] = updated_location
    updated_location_to_count[updated_location] += 1

updated_location_to_count = {k: v for k, v in sorted(updated_location_to_count.items(), 
                         key=lambda item: item[1],
                         reverse=True)}

updated_locations = set(file_name_to_updated_location.values())

print('Found {} updated locations in the original metadata:'.format(len(updated_locations)))
for loc in updated_location_to_count:
    print('{}: {}'.format(loc,updated_location_to_count[loc]))


#%% Re-write metadata

for im in d['images']:
    im['location'] = file_name_to_updated_location[im['file_name']]
d['info']['version'] = 1.14

with open(output_file,'w') as f:
    json.dump(d,f,indent=1)
    

#%% For each location, sample some random images to make sure they look consistent

input_base = r'd:\lila\nacti-unzipped'
assert os.path.isdir(input_base)

location_to_images = defaultdict(list)

for im in d['images']:
    location_to_images[im['location']].append(im)
    
n_to_sample = 10
import random
random.seed(0)
sampling_folder_base = r'g:\temp\nacti_samples'

for location in tqdm(location_to_images):
    
    images_this_location = location_to_images[location]
    if len(images_this_location) > n_to_sample:
        images_this_location = random.sample(images_this_location,n_to_sample)
        
    for i_image,im in enumerate(images_this_location):
        
        fn_relative = im['file_name']
        source_fn_abs = os.path.join(input_base,fn_relative)
        assert os.path.isfile(source_fn_abs)
        ext = os.path.splitext(fn_relative)[1]            
        target_fn_abs = os.path.join(sampling_folder_base,'{}/{}'.format(
            location,'image_{}{}'.format(str(i_image).zfill(2),ext)))
        os.makedirs(os.path.dirname(target_fn_abs),exist_ok=True)
        shutil.copyfile(source_fn_abs,target_fn_abs)
        
    # ...for each image
    
# ...for each location
            



"""
 
 prepare_zsl_imerit.py

 Prepare ZSL Borneo data for annotation (convert input data to iMerit-friendly format).

"""

#%% Imports and constants

import json
import os

from tqdm import tqdm
from operator import itemgetter
from shutil import copyfile

from md_visualization import visualize_db
from data_management.databases import integrity_check_json_db
from data_management.cct_json_utils import IndexedJsonDb

annotation_list_filename = r'd:\wildlife_data\zsl_borneo\all_img_ids_to_bbox.json'
image_json = r'd:\wildlife_data\zsl_borneo\201906cameratraps\0.5\zsl_camera_traps_201906.json'
image_base = r'd:\wildlife_data\zsl_borneo\201906cameratraps\0.5'
output_base = r'd:\wildlife_data\zsl_borneo'

human_classes = ['human','hunter']


#%% Load data

with open(annotation_list_filename,'r') as f:
    annotation_list = json.load(f)
    
# with open(image_json,'r') as f:
#    data = json.load(f)
indexedData = IndexedJsonDb(image_json)

print('Done loading data')    
    

#%% Validate data

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = image_base
options.bCheckImageSizes = False
options.bCheckImageExistence = True
options.bFindUnusedImages = False
    
sortedCategories = integrity_check_json_db.integrity_check_json_db(indexedData.db,options)


#%% Label previews

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 500
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = True
viz_options.sort_by_filename = False
html_output_file,image_db = visualize_db.visualize_db(indexedData.db,
                                                        os.path.join(output_base,'preview'),
                                                        image_base,viz_options)
os.startfile(html_output_file)


#%% Collect images to annotate

images_to_annotate = []

annotation_list = set(annotation_list)
n_humans = 0

for im in tqdm(indexedData.db['images']):
    
    class_names = indexedData.get_classes_for_image(im)
    b_human = False
    for cn in class_names:
        if cn.lower() in human_classes:
            b_human = True
            n_humans += 1
            break
        
    if b_human or im['id'] in annotation_list:
        images_to_annotate.append(im)
    

print('Found {} of {} images ({} humans)'.format(len(images_to_annotate),len(annotation_list),n_humans))
assert len(images_to_annotate) >= len(annotation_list)



#%% Sort by sequence and frame

images_to_annotate = sorted(images_to_annotate, key=itemgetter('seq_id', 'frame_num')) 


#%% Copy to a folder by GUID

# dataset[dataset_id].seq[sequence_id].frame[frame_number].img[img_id].extension

imerit_output_base = os.path.join(output_base,'imerit_batch_9')
os.makedirs(imerit_output_base,exist_ok=True)

# im = images_to_annotate[0]
for im in tqdm(images_to_annotate):
    
    relative_path = im['file_name']
    extension = os.path.splitext(relative_path)[1]
    frame_num = im['frame_num']
    seq_id = im['seq_id']
    id = im['id']
    assert '.' not in id
    input_full_path = os.path.join(image_base,relative_path)
    assert os.path.isfile(input_full_path)
    output_filename = 'datasetzslborneo.seq' + '{0:0>8d}'.format(seq_id) + '.frame' + \
        '{0:0>4d}'.format(frame_num) + '.img' + id + extension
    im['imerit_filename'] = output_filename
    output_full_path = os.path.join(imerit_output_base,output_filename)
    assert not os.path.isfile(output_full_path)
    copyfile(input_full_path,output_full_path)

# ...for each image
    

#%% Write out the annotation list

imerit_batch9_json_filename = os.path.join(imerit_output_base,'imerit_batch_9.json')
with open(imerit_batch9_json_filename,'w') as f:
    json.dump(images_to_annotate, f, indent=2)
    
    


"""

 ubc_to_json.py

 Convert the .csv file provided for the UBC data set to a 
 COCO-camera-traps .json file

 Images were provided in eight folders, each of which contained a .csv
 file with annotations.  Those annotations came in two slightly different 
 formats, the two formats corresponding to folders starting with "SC_" and 
 otherwise.

"""

#%% Constants and environment

import pandas as pd
import os
import json
import uuid
import numpy as np
import shutil

from tqdm import tqdm
from PIL import Image

from md_visualization import visualize_db
from data_management.databases import integrity_check_json_db
from md_utils.path_utils import find_images

input_base = r'e:\ubc'
assert(os.path.isdir(input_base))

output_base = r'f:\data_staging\ubc'
output_json_file = os.path.join(output_base,'ubc.json')
file_list_file = os.path.join(output_base,'all_files.txt')

os.makedirs(output_base,exist_ok=True)

# Map Excel column names - which vary a little across spreadsheets - to a common set of names
mapped_fields = {"Survey.Name" : "survey_name",
                 "project_id": "survey_name",
                 "Camera.Name": "camera_name",
                 "station_id": "camera_name",
                 "Media.Filename": "filename",
                 "orig_file": "filename",
                 "timestamp_pst": "datetime",
                 "Date.Time": "datetime",
                 "Species": "species",
                 "latin_name": "species",
                 "common.name": "common_name",
                 "common_names": "common_name",
                 "Sighting.Quantity": "species_count"
                }

category_mappings = {
    'bird_spp.':'unknown_bird',
    'dog_dog':'dog',
    'hiker_hiker':'hiker',
    'quad_quad':'quad',
    'skier_skier':'skier',
    'snowshoer_snowshoer':'showshoer',
    'quad_quad':'quad'
    }

target_fields = ['species_count','group_count','behaviour']


#%% Enumerate images

# Load from file if we've already enumerated
if os.path.isfile(file_list_file):
    with open(file_list_file,'r') as f:
        files = f.readlines()
    files = [s.strip() for s in files]
    image_full_paths = files
    print('Loaded {} images from {}'.format(len(image_full_paths),file_list_file))
else:
    image_full_paths = find_images(input_base, recursive=True)
    with open(file_list_file,'w') as f:
        for line in image_full_paths:
            f.write(line + '\n')        
    print('Enumerated {} images from {}'.format(len(image_full_paths),input_base))
    
image_full_paths_set = set(image_full_paths)
image_relative_paths = [os.path.relpath(fn,input_base) for fn in image_full_paths]
image_relative_paths_set = set(image_relative_paths)


#%% Create CCT dictionaries

annotations = []
image_ids_to_images = {}
category_name_to_category = {}

# Force the empty category to be ID 0
empty_category = {}
empty_category['name'] = 'empty'
empty_category['id'] = 0
empty_category['common_name'] = 'empty'
category_name_to_category['empty'] = empty_category
next_category_id = 1

latin_to_common = {}

folders = os.listdir(input_base)

# To simplify debugging of the loop below
i_folder = 0; folder = folders[i_folder]; 


##%% Create CCT dictionaries (loop)

invalid_images = []

for i_folder,folder in enumerate(folders):
        
    ##%% 
    
    print('\nProcessing folder {} of {}: {}'.format(i_folder,len(folders),folder))
    
    filenames_to_rows = {}
    filenames_with_multiple_annotations = []
    missing_images = []
    image_directory = os.path.join(input_base, folder)
    files = os.listdir(image_directory)
    files = list(filter(lambda f: f.endswith('.csv'), files))
    input_metadata_file = os.path.join(input_base, folder, files[0])
    assert(os.path.isfile(input_metadata_file))
    
    # Read source data for this folder
    input_metadata = pd.read_csv(input_metadata_file)
    
    # Rename columns
    input_metadata.rename(columns=mapped_fields, inplace=True)
    print('Read {} columns and {} rows from metadata file'.format(len(input_metadata.columns), len(input_metadata)))
    
    if folder.startswith("SC_"):
        # Folder name is the first two characters of the filename
        #
        # Create relative path names from the filename itself
        input_metadata['image_relative_path'] = input_metadata['filename'].apply(
            lambda x: os.path.join(folder, x[0:2], x.replace(".JPG", ".jpg")))
    else:
        # Folder name is the camera name
        #
        # Create relative path names from camera name and filename    
        input_metadata['image_relative_path'] = input_metadata[['camera_name', 'filename']].apply(
            lambda x: os.path.join(folder, x[0], x[1]), axis = 1)

    # Which of our images are in the spreadsheet?
    # i_row = 0; fn = input_metadata['image_relative_path'][i_row]        
    for i_row, image_relative_path in enumerate(input_metadata['image_relative_path']):
        
        if image_relative_path in filenames_to_rows:
            filenames_with_multiple_annotations.append(image_relative_path)
            filenames_to_rows[image_relative_path].append(i_row)
        else:
            filenames_to_rows[image_relative_path] = [i_row]
            image_full_path = os.path.join(input_base, image_relative_path)
            
            if not image_full_path in image_full_paths_set:
                assert image_relative_path not in image_relative_paths_set
                missing_images.append(image_full_path)
        
    print('Finished verifying image existence for {} files in {} rows\nFound {} filenames with multiple labels, {} missing images'.format(
        len(input_metadata), len(filenames_to_rows), 
        len(filenames_with_multiple_annotations), len(missing_images)))
    
    
    ##%% Check for images that aren't included in the metadata file

    # Find all the images in this folder
    image_relative_paths_this_folder = [s for s in image_relative_paths if s.startswith(folder)]

    # Which of these aren't in the spreadsheet?
    annotated_files_this_folder = list(filenames_to_rows.keys())
    annotated_files_this_folder_set = set(annotated_files_this_folder)
    unannotated_images = [s for s in image_relative_paths_this_folder if s not in annotated_files_this_folder_set]

    print('Found {} unannotated images (of {}) in this folder'.format(
            len(unannotated_images),len(image_relative_paths_this_folder)))
    
    
    ##%% Create entries in CCT dictionaries
    
    image_relative_path = list(filenames_to_rows.keys())[0]
    
    for image_relative_path in list(filenames_to_rows.keys()):
        
        # Only process images we have on disk
        if image_relative_path not in image_relative_paths_set:
            continue
        
        image_full_path = os.path.join(input_base,image_relative_path)
        
        # This is redundant, but doing this for clarity, at basically no performance
        # cost since we need to *read* the images below to check validity.
        assert os.path.isfile(image_full_path)
        
        img_id = image_relative_path.replace('\\','/').replace('/','_').replace(' ','_')
        row_indices = filenames_to_rows[image_relative_path]
        
        # i_row = row_indices[0]
        for i_row in row_indices:
            
            row = input_metadata.iloc[i_row]
            assert(row['image_relative_path'] == image_relative_path)
            timestamp = row['datetime']            
            location = row['survey_name'] + '_' + row['camera_name']
            
            if img_id in image_ids_to_images:
                im = image_ids_to_images[img_id]
                assert im['file_name'] == image_relative_path
                assert im['location'] == location
            else:
                im = {}
                
                try:
                    pil_image = Image.open(image_full_path)        
                    width, height = pil_image.size
                    im['width'] = width
                    im['height'] = height
                except:
                    # These generally represent zero-byte images in this data set, don't try
                    # to find the very small handful that might be other kinds of failures we 
                    # might want to keep around.
                    # print('Error opening image {}'.format(image_relative_path))
                    invalid_images.append(image_relative_path)
                    continue
                
                im['id'] = img_id
                im['file_name'] = image_relative_path
                im['datetime'] = timestamp
                im['location'] = location

                image_ids_to_images[img_id] = im

            species = row['species'].lower().strip().replace(' ','_')
            
            if (isinstance(species,float) or \
                (isinstance(species,str) and (len(species) == 0))):
                category_name = 'empty'
            else:
                category_name = species
            del species
            
            category_name = category_name.strip().lower()
            
            common_name = row['common_name']
            if isinstance(common_name,float) and np.isnan(common_name):
                common_name = ''
            else:
                common_name = str(common_name).lower().strip().replace(', ',',').replace(' ','_')
            
            for k,v in category_mappings.items():
                common_name = common_name.replace(k,v)
                category_name = category_name.replace(k,v)
            common_name = common_name.replace('.','').replace('spp','species')
            category_name = category_name.replace('.','').replace('spp','species')
                        
            if category_name == 'passerine_species' and common_name != '' and common_name != 'passerine_species':
                category_name = common_name
                
            # If we've seen this category before...
            if category_name in category_name_to_category:
                
                category = category_name_to_category[category_name]
                category_id = category['id'] 
                
                # ...make sure it used the same latin --> common mapping
                #
                # If the previous instance had no mapping, use the new one.
                if category['common_name'] == '':
                    category['common_name'] = common_name
                else:
                    # assert common_name == category['common_name']
                    if common_name != category['common_name']:
                        print('Warning: common name {} used for species {}, previously {}'.format(
                            common_name,category_name,category['common_name']))
                    
            else:
                
                category_id = next_category_id
                category = {}
                category['id'] = category_id
                category['name'] = category_name
                category['common_name'] = common_name
                category_name_to_category[category_name] = category
                next_category_id += 1
            
            # Create an annotation
            ann = {}        
            ann['id'] = str(uuid.uuid1())
            ann['image_id'] = im['id']    
            ann['category_id'] = category_id
                        
            for target_field in target_fields:
                if target_field in input_metadata.columns:
                    val = row[target_field]
                    if isinstance(val,float) and np.isnan(val):
                        val = ''
                    else:
                        val = str(val).strip()
                    ann[target_field] = val
                
            annotations.append(ann)

        # ...for each annotation we found for this image
            
    # ...for each image
             
# ...for each dataset
            
images = list(image_ids_to_images.values())
categories = list(category_name_to_category.values())

# Print all of our species mappings
for c in categories:
    print(c['name'].ljust(30) + c['common_name'])

print('Finished creating CCT dictionaries, loaded {} images of {} total on disk ({} invalid)'.format(
    len(images), len(image_relative_paths_set), len(invalid_images)))


#%% Copy images for which we actually have annotations to a new folder, lowercase everything

# im = images[0]
for im in tqdm(images):
    relative_filename = im['file_name']
    input_filename = os.path.join(input_base,relative_filename)
    output_filename = os.path.join(output_base,relative_filename).lower()
    os.makedirs(os.path.dirname(output_filename),exist_ok=True)
    
    shutil.copy(input_filename, output_filename)
    im['file_name'] = im['file_name'].lower()
    im['id'] = im['id'].lower()
    

#%% Create info struct

info = {}
info['year'] = 2020
info['version'] = 1
info['description'] = 'UBC Camera Traps'
info['contributor'] = 'UBC'


#%% Convert image IDs to lowercase in annotations, tag as sequence level

# While there isn't any sequence information, the nature of false positives
# here leads me to believe the images were labeled at the sequence level, so
# we should trust labels more when positives are verified.  Overall false
# positive rate looks to be between 1% and 5%.
    
for ann in annotations:
    ann['image_id'] = ann['image_id'].lower()
    ann['sequence_level_annotation'] = True


#%% Write output

json_data = {}
json_data['images'] = images
json_data['annotations'] = annotations
json_data['categories'] = categories
json_data['info'] = info
json.dump(json_data, open(output_json_file, 'w'), indent=2)

print('Finished writing .json file with {} images, {} annotations, and {} categories'.format(
    len(images), len(annotations), len(categories)))


#%% Validate output

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = output_base
options.bCheckImageSizes = False
options.bCheckImageExistence = False
options.bFindUnusedImages = True

sortedCategories, data, errors = integrity_check_json_db.integrity_check_json_db(
    output_json_file, options)


#%% Preview labels

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 2000
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = True
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
html_output_file, image_db = visualize_db.visualize_db(db_path=output_json_file,
                                                         output_dir=os.path.join(
                                                             output_base, 'preview'),
                                                         image_base_dir=output_base,
                                                         options=viz_options)
os.startfile(html_output_file)



r"""

find_repeat_detections.py

If you want to use this script, we recommend that you read the RDE user's guide:

https://github.com/agentmorris/MegaDetector/tree/master/api/batch_processing/postprocessing/repeat_detection_elimination

Really, don't try to run this script without reading the user's guide, you'll think 
it's more magical than it is. 

This script looks through a sequence of detections in the API output json file, and finds 
candidates that might be "repeated false positives", i.e. that random branch that the 
detector thinks is an animal/person/vehicle.

Typically after running this script, you would do a manual step to remove 
true positives, then run remove_repeat_detections to produce a final output file.

There's no way that statement was self-explanatory; see the user's guide.

This script is just a command-line driver for repeat_detections_core.py.

"""

#%% Constants and imports

import argparse
import os
import sys

from md_utils import ct_utils
from api.batch_processing.postprocessing.repeat_detection_elimination import repeat_detections_core


#%% Interactive driver

if False:
    
    #%%

    baseDir = ''

    options = repeat_detections_core.RepeatDetectionOptions()
    options.bRenderHtml = True
    options.imageBase = baseDir
    options.outputBase = os.path.join(baseDir, 'repeat_detections')
    options.filenameReplacements = {}  # E.g., {'20190430cameratraps\\':''}

    options.confidenceMin = 0.15
    options.confidenceMax = 1.01
    options.iouThreshold = 0.85
    options.occurrenceThreshold = 8
    options.maxSuspiciousDetectionSize = 0.2

    options.filterFileToLoad = ''
    options.filterFileToLoad = os.path.join(baseDir,r'..\detectionIndex.json')

    options.debugMaxDir = -1
    options.debugMaxRenderDir = -1
    options.debugMaxRenderDetection = -1
    options.debugMaxRenderInstance = -1

    options.bParallelizeComparisons = False
    options.bParallelizeRendering = False
    options.excludeClasses = [2]

    inputFilename = os.path.join(baseDir, 'blah.json')
    outputFilename = os.path.join(baseDir, 'blah.json')

    results = repeat_detections_core.find_repeat_detections(inputFilename, outputFilename, options)


#%% Command-line driver

def main():
    
    defaultOptions = repeat_detections_core.RepeatDetectionOptions()

    parser = argparse.ArgumentParser()
    
    parser.add_argument('inputFile', type=str, help='MD results .json file to process')
    
    parser.add_argument('--outputFile', action='store', type=str, default=None,
                        help='.json file to write filtered results to... do not use this if you are going to ' + \
                             'do manual review of the repeat detection images (which you should)')
        
    parser.add_argument('--imageBase', action='store', type=str, default='',
                        help='Image base dir, relevant if renderHtml is True or if ' + \
                             '"omitFilteringFolder" is not set')
                            
    parser.add_argument('--outputBase', action='store', type=str, default='',
                        help='HTML or filtering folder output dir')
    
    parser.add_argument('--confidenceMin', action='store', type=float,
                        default=defaultOptions.confidenceMin,
                        help='Detection confidence threshold; don\'t process anything below this')
    
    parser.add_argument('--confidenceMax', action='store', type=float,
                        default=defaultOptions.confidenceMax,
                        help='Detection confidence threshold; don\'t process anything above this')
    
    parser.add_argument('--iouThreshold', action='store', type=float,
                        default=defaultOptions.iouThreshold,
                        help='Detections with IOUs greater than this are considered ' + \
                             '"the same detection"')
        
    parser.add_argument('--occurrenceThreshold', action='store', type=int,
                        default=defaultOptions.occurrenceThreshold,
                        help='More than this many near-identical detections in a group ' + \
                             '(e.g. a folder) is considered suspicious')
        
    parser.add_argument('--minSuspiciousDetectionSize', action='store', type=float,
                        default=defaultOptions.minSuspiciousDetectionSize,
                        help='Detections smaller than this fraction of image area are not ' + \
                             'considered suspicious')

    parser.add_argument('--maxSuspiciousDetectionSize', action='store', type=float,
                        default=defaultOptions.maxSuspiciousDetectionSize,
                        help='Detections larger than this fraction of image area are not ' + \
                             'considered suspicious')

    parser.add_argument('--maxImagesPerFolder', action='store', type=int,
                        default=defaultOptions.maxImagesPerFolder,
                        help='Ignore folders with more than this many images in them')
    
    parser.add_argument('--excludeClasses', action='store', nargs='+', type=int,
                        default=None,
                        help='List of integer classes we don\'t want to treat as suspicious, separated by spaces.')
    
    parser.add_argument('--pass_detections_to_processes_method', action='store', type=str,
                        default=defaultOptions.pass_detections_to_processes_method,
                        help='Pass detections information to/from workers via "memory" (default) or "files"')
    
    parser.add_argument('--nWorkers', action='store', type=int,
                        default=defaultOptions.nWorkers,
                        help='Level of parallelism for rendering and IOU computation')
    
    parser.add_argument('--parallelizationUsesProcesses', action='store_false', 
                        dest='parallelizationUsesThreads',
                        help='Parallelize with processes (defaults to threads)')    
    
    parser.add_argument('--filterFileToLoad', action='store', type=str, default='',  
                        help='Path to detectionIndex.json, which should be inside a ' + \
                             'folder of images that are manually verified to _not_ ' + \
                             'contain valid animals')

    parser.add_argument('--omitFilteringFolder', action='store_false',
                        dest='bWriteFilteringFolder',
                        help='Should we create a folder of rendered detections for post-filtering?')
    
    parser.add_argument('--debugMaxDir', action='store', type=int, default=-1, 
                        help='For debugging only, limit the number of directories we process')
    parser.add_argument('--debugMaxRenderDir', action='store', type=int, default=-1,
                        help='For debugging only, limit the number of directories we render')
    parser.add_argument('--debugMaxRenderDetection', action='store', type=int, default=-1,
                        help='For debugging only, limit the number of detections we process per folder')
    parser.add_argument('--debugMaxRenderInstance', action='store', type=int, default=-1,
                        help='For debugging only, limit the number of instances we process per detection')

    parser.add_argument('--forceSerialComparisons', action='store_false',
                        dest='bParallelizeComparisons',
                        help='Disable parallelization during the comparison stage')
    parser.add_argument('--forceSerialRendering', action='store_false',
                        dest='bParallelizeRendering',
                        help='Disable parallelization during the rendering stage')
    
    parser.add_argument('--maxOutputImageWidth', action='store', type=int,
                        default=defaultOptions.maxOutputImageWidth,
                        help='Maximum output size for thumbnail images') 
    
    parser.add_argument('--lineThickness', action='store', type=int,
                        default=defaultOptions.lineThickness,
                        help='Line thickness thumbnail images')    
    
    parser.add_argument('--boxExpansion', action='store', type=int,
                        default=defaultOptions.boxExpansion,
                        help='Box expansion for thumbnail images')
        
    parser.add_argument('--nDirLevelsFromLeaf', type=int,
                        default=defaultOptions.nDirLevelsFromLeaf,
                        help='Number of levels from the leaf folders to use for repeat ' + \
                             'detection (0 == leaves)')

    parser.add_argument('--bRenderOtherDetections', action='store_true',
                        help='Show non-target detections in light gray on each image')
    
    parser.add_argument('--bRenderDetectionTiles', action='store_true',
                        help='Should we render a grid showing every instance (up to a limit) for each detection?')
    
    parser.add_argument('--detectionTilesPrimaryImageWidth', type=int,
                        default=defaultOptions.detectionTilesPrimaryImageWidth,
                        help='The width of the main image when rendering images with detection tiles')

    parser.add_argument('--renderHtml', action='store_true',
                        dest='bRenderHtml', help='Should we render HTML output?')
    
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    # Convert to an options object
    options = repeat_detections_core.RepeatDetectionOptions()

    ct_utils.args_to_object(args, options)

    repeat_detections_core.find_repeat_detections(args.inputFile, args.outputFile, options)

if __name__ == '__main__':
    main()



"""

 wellington_to_json.py

 Convert the .csv file provided for the Wellington data set to a 
 COCO-camera-traps .json file

"""

#%% Constants and environment

import pandas as pd
import os
import glob
import json
import re
import uuid
import time
import ntpath
import humanfriendly
import PIL

from tqdm import tqdm

input_metadata_file = os.path.expanduser('~/data/wct/wellington_camera_traps.csv')
output_file = os.path.expanduser('~/data/wct/wellington_camera_traps.json')
image_directory = os.path.expanduser('~/data/wct/images')
preview_dir = os.path.expanduser('~/data/wct/preview')

assert(os.path.isdir(image_directory))


#%% Read source data

input_metadata = pd.read_csv(input_metadata_file)

print('Read {} columns and {} rows from metadata file'.format(len(input_metadata.columns),
      len(input_metadata)))

# Filenames were provided as *.jpg, but images were *.JPG, converting here
input_metadata['file'] = input_metadata['file'].apply(lambda x: x.replace('.jpg','.JPG'))

print('Converted extensions to uppercase')


#%% Map filenames to rows, verify image existence

# Takes ~30 seconds, since it's checking the existence of ~270k images

start_time = time.time()
filenames_to_rows = {}
image_filenames = input_metadata.file

duplicate_rows = []

# Build up a map from filenames to a list of rows, checking image existence as we go
for i_file,fn in enumerate(image_filenames):
    
    if (fn in filenames_to_rows):
        duplicate_rows.append(i_file)
        filenames_to_rows[fn].append(i_file)
    else:
        filenames_to_rows[fn] = [i_file]
        image_path = os.path.join(image_directory,fn)
        assert(os.path.isfile(image_path))

elapsed = time.time() - start_time
print('Finished verifying image existence in {}, found {} filenames with multiple labels'.format(
      humanfriendly.format_timespan(elapsed),len(duplicate_rows)))

# I didn't expect this to be true a priori, but it appears to be true, and
# it saves us the trouble of checking consistency across multiple occurrences
# of an image.
assert(len(duplicate_rows) == 0)    
    
    
#%% Check for images that aren't included in the metadata file

# Enumerate all images
image_full_paths = glob.glob(os.path.join(image_directory,'*.JPG'))

for i_image,image_path in enumerate(image_full_paths):
    
    fn = ntpath.basename(image_path)
    assert(fn in filenames_to_rows)

print('Finished checking {} images to make sure they\'re in the metadata'.format(
        len(image_full_paths)))


#%% Create CCT dictionaries

# Also gets image sizes, so this takes ~6 minutes
#
# Implicitly checks images for overt corruptness, i.e. by not crashing.

images = []
annotations = []

# Map categories to integer IDs (that's what COCO likes)
next_category_id = 0
categories_to_category_id = {}
categories_to_counts = {}

# For each image
#
# Because in practice images are 1:1 with annotations in this data set,
# this is also a loop over annotations.

start_time = time.time()

sequence_frame_ids = set()

# image_name = image_filenames[0]
for image_name in tqdm(image_filenames):
    
    rows = filenames_to_rows[image_name]
    
    # As per above, this is convenient and appears to be true; asserting to be safe
    assert(len(rows) == 1)    
    i_row = rows[0]
    
    row = input_metadata.iloc[i_row]
    
    im = {}
    # Filenames look like "290716114012001a1116.jpg"
    im['id'] = image_name.split('.')[0]    
    im['file_name'] = image_name
    
    # This gets imported as an int64
    im['seq_id'] = str(row['sequence'])
        
    # These appear as "image1", "image2", etc.
    frame_id = row['image_sequence']
    m = re.match('^image(\d+)$',frame_id)
    assert (m is not None)
    im['frame_num'] = int(m.group(1))-1
    
    # Make sure we haven't seen this sequence before
    sequence_frame_id = im['seq_id'] + '_' + str(im['frame_num'])
    assert sequence_frame_id not in sequence_frame_ids
    sequence_frame_ids.add(sequence_frame_id)
    
    # In the form "001a"
    im['location'] = row['site']
    
    # Can be in the form '111' or 's46'
    im['camera'] = row['camera']
    
    # In the form "7/29/2016 11:40"
    im['datetime'] = row['date']
    
    # Check image height and width
    image_path = os.path.join(image_directory,fn)
    assert(os.path.isfile(image_path))
    pil_image = PIL.Image.open(image_path)
    width, height = pil_image.size
    im['width'] = width
    im['height'] = height

    images.append(im)
    
    category = row['label'].lower()
    
    # Use 'empty', to be consistent with other data on lila    
    if (category == 'nothinghere'):
        category = 'empty'
        
    # Have we seen this category before?
    if category in categories_to_category_id:
        category_id = categories_to_category_id[category]
        categories_to_counts[category] += 1
    else:
        category_id = next_category_id
        categories_to_category_id[category] = category_id
        categories_to_counts[category] = 0
        next_category_id += 1
    
    # Create an annotation
    ann = {}
    
    # The Internet tells me this guarantees uniqueness to a reasonable extent, even
    # beyond the sheer improbability of collisions.
    ann['id'] = str(uuid.uuid1())
    ann['image_id'] = im['id']    
    ann['category_id'] = category_id
    
    annotations.append(ann)
    
# ...for each image
    
# Convert categories to a CCT-style dictionary

categories = []

for category in categories_to_counts:
    print('Category {}, count {}'.format(category,categories_to_counts[category]))
    category_id = categories_to_category_id[category]
    cat = {}
    cat['name'] = category
    cat['id'] = category_id
    categories.append(cat)
    
elapsed = time.time() - start_time
print('Finished creating CCT dictionaries in {}'.format(
      humanfriendly.format_timespan(elapsed)))
    

#%% Create info struct

info = {}
info['year'] = 2018
info['version'] = '1.01'
info['description'] = 'Wellington Camera Traps'
info['secondary_contributor'] = 'Converted to COCO .json by Dan Morris'
info['contributor'] = 'Victor Anton'


#%% Write output

json_data = {}
json_data['images'] = images
json_data['annotations'] = annotations
json_data['categories'] = categories
json_data['info'] = info
json.dump(json_data,open(output_file,'w'),indent=1)

print('Finished writing .json file with {} images, {} annotations, and {} categories'.format(
        len(images),len(annotations),len(categories)))


#%% Validate .json files

from data_management.databases import integrity_check_json_db

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = image_directory
options.bCheckImageSizes = False
options.bCheckImageExistence = True
options.bFindUnusedImages = True

sorted_categories, data, error_info = integrity_check_json_db.integrity_check_json_db(output_file, options)


#%% Preview labels

from md_visualization import visualize_db

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 2000
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = False
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
viz_options.classes_to_exclude = ['test']
html_output_file, image_db = visualize_db.visualize_db(db_path=output_file,
                                                         output_dir=os.path.join(
                                                         preview_dir),
                                                         image_base_dir=image_directory,
                                                         options=viz_options)

from md_utils import path_utils
path_utils.open_file(html_output_file)



"""

process_video.py

Splits a video (or folder of videos) into frames, runs the frames through run_detector_batch.py,
and optionally stitches together results into a new video with detection boxes.

Operates by separating the video into frames, typically sampling every Nth frame, and writing
those frames to disk, before running MD.  This approach clearly has a downside: it requires
a bunch more disk space, compared to extracting frames and running MD on them without ever
writing them to disk.  The upside, though, is that this approach allows you to run repeat
detection elimination after running MegaDetector, and it allows allows more efficient re-use
of frames if you end up running MD more than once, or running multiple versions of MD.

TODO: optionally skip writing frames to disk, and process frames in memory.

"""

#%% Imports

import os
import sys
import tempfile
import argparse
import itertools
import json
import shutil

from detection import run_detector_batch
from md_visualization import visualize_detector_output
from md_utils.ct_utils import args_to_object
from detection.video_utils import video_to_frames
from detection.video_utils import frames_to_video
from detection.video_utils import frame_results_to_video_results
from detection.video_utils import video_folder_to_frames
from uuid import uuid1

from detection.video_utils import default_fourcc


#%% Classes

class ProcessVideoOptions:
    """
    Options controlling the behavior of process_video()
    """
    
    #: Can be a model filename (.pt or .pb) or a model name (e.g. "MDV5A")
    model_file = 'MDV5A'
    
    #: Video (of folder of videos) to process
    input_video_file = ''

    #: .json file to which we should write results
    output_json_file = None
    
    #: File to which we should write a video with boxes, only relevant if 
    #: render_output_video is True
    output_video_file = None
    
    #: Folder to use for extracted frames; will use a folder in system temp space
    #: if this is None
    frame_folder = None
    
    # Folder to use for rendered frames (if rendering output video); will use a folder 
    #: in system temp space if this is None
    frame_rendering_folder = None
    
    #: Should we render a video with detection boxes?
    #:
    #: Only supported when processing a single video, not a folder.
    render_output_video = False
    
    #: If we are rendering boxes to a new video, should we keep the temporary
    #: rendered frames?
    keep_rendered_frames = False
    
    #: Should we keep the extracted frames?
    keep_extracted_frames = False
    
    #: Should we delete the entire folder the extracted frames are written to?
    #:
    #: By default, we delete the frame files but leave the (probably-empty) folder in place, 
    #: for no reason other than being paranoid about deleting folders.
    force_extracted_frame_folder_deletion = False
    
    #: Should we delete the entire folder the rendered frames are written to?
    #:
    #: By default, we delete the frame files but leave the (probably-empty) folder in place,
    #: for no reason other than being paranoid about deleting folders.
    force_rendered_frame_folder_deletion = False
     
    #: If we've already run MegaDetector on this video or folder of videos, i.e. if we 
    #: find a corresponding MD results file, should we re-use it?  Defaults to reprocessing.
    reuse_results_if_available = False
    
    #: If we've already split this video or folder of videos into frames, should we 
    #: we re-use those extracted frames?  Defaults to reprocessing.
    reuse_frames_if_available = False
    
    #: If [input_video_file] is a folder, should we search for videos recursively?
    recursive = False 
    
    #: Enable additional debug console output
    verbose = False
    
    #: fourcc code to use for writing videos; only relevant if render_output_video is True
    fourcc = None

    #: Confidence threshold to use for writing videos with boxes, only relevant if
    #: if render_output_video is True.  Defaults to choosing a reasonable threshold
    #: based on the model version.
    rendering_confidence_threshold = None
    
    #: Detections below this threshold will not be included in the output file.
    json_confidence_threshold = 0.005
    
    #: Sample every Nth frame; set to None (default) or 1 to sample every frame.  Typically
    #: we sample down to around 3 fps, so for typical 30 fps videos, frame_sample=10 is a 
    #: typical value.
    frame_sample = None
    
    #: Number of workers to use for parallelization; set to <= 1 to disable parallelization
    n_cores = 1

    #: For debugging only, stop processing after a certain number of frames.
    debug_max_frames = -1
    
    #: File containing non-standard categories, typically only used if you're running a non-MD
    #: detector.
    class_mapping_filename = None

# ...class ProcessVideoOptions


#%% Functions

def process_video(options):
    """
    Process a single video through MD, optionally writing a new video with boxes
    
    Args: 
        options (ProcessVideoOptions): all the parameters used to control this process,
            including filenames; see ProcessVideoOptions for details
            
    Returns:
        dict: frame-level MegaDetector results, identical to what's in the output .json file
    """

    if options.output_json_file is None:
        options.output_json_file = options.input_video_file + '.json'

    if options.render_output_video and (options.output_video_file is None):
        options.output_video_file = options.input_video_file + '.detections.mp4'

    tempdir = os.path.join(tempfile.gettempdir(), 'process_camera_trap_video')    
    os.makedirs(tempdir,exist_ok=True)
    
    # TODO:
    # 
    # This is a lazy fix to an issue... if multiple users run this script, the
    # "process_camera_trap_video" folder is owned by the first person who creates it, and others
    # can't write to it.  I could create uniquely-named folders, but I philosophically prefer
    # to put all the individual UUID-named folders within a larger folder, so as to be a 
    # good tempdir citizen.  So, the lazy fix is to make this world-writable.
    try:
        os.chmod(tempdir,0o777)
    except Exception:
        pass
    
    if options.frame_folder is not None:
        frame_output_folder = options.frame_folder
    else:
        frame_output_folder = os.path.join(
            tempdir, os.path.basename(options.input_video_file) + '_frames_' + str(uuid1()))
        
    # TODO: keep track of whether we created this folder, delete if we're deleting the extracted
    # frames and we created the folder, and the output files aren't in the same folder.  For now,
    # we're just deleting the extracted frames and leaving the empty folder around in this case.
    os.makedirs(frame_output_folder, exist_ok=True)

    frame_filenames, Fs = video_to_frames(
        options.input_video_file, frame_output_folder, 
        every_n_frames=options.frame_sample, overwrite=(not options.reuse_frames_if_available))

    image_file_names = frame_filenames
    if options.debug_max_frames > 0:
        image_file_names = image_file_names[0:options.debug_max_frames]

    if options.reuse_results_if_available and \
        os.path.isfile(options.output_json_file):
            print('Loading results from {}'.format(options.output_json_file))
            with open(options.output_json_file,'r') as f:
                results = json.load(f)
    else:
        results = run_detector_batch.load_and_run_detector_batch(
            options.model_file, image_file_names,
            confidence_threshold=options.json_confidence_threshold,
            n_cores=options.n_cores,
            quiet=(not options.verbose),
            class_mapping_filename=options.class_mapping_filename)
    
        run_detector_batch.write_results_to_file(
            results, options.output_json_file,
            relative_path_base=frame_output_folder,
            detector_file=options.model_file,
            custom_metadata={'video_frame_rate':Fs})

    
    ## (Optionally) render output video
    
    if options.render_output_video:
        
        # Render detections to images
        if options.frame_rendering_folder is not None:
            rendering_output_dir = options.frame_rendering_folder
        else:
            rendering_output_dir = os.path.join(
                tempdir, os.path.basename(options.input_video_file) + '_detections')
            
        # TODO: keep track of whether we created this folder, delete if we're deleting the rendered
        # frames and we created the folder, and the output files aren't in the same folder.  For now,
        # we're just deleting the rendered frames and leaving the empty folder around in this case.
        os.makedirs(rendering_output_dir,exist_ok=True)
        
        detected_frame_files = visualize_detector_output.visualize_detector_output(
            detector_output_path=options.output_json_file,
            out_dir=rendering_output_dir,
            images_dir=frame_output_folder,
            confidence_threshold=options.rendering_confidence_threshold)

        # Combine into a video
        if options.frame_sample is None:
            rendering_fs = Fs
        else:
            rendering_fs = Fs / options.frame_sample
            
        print('Rendering video to {} at {} fps (original video {} fps)'.format(
            options.output_video_file,rendering_fs,Fs))
        frames_to_video(detected_frame_files, rendering_fs, options.output_video_file, codec_spec=options.fourcc)
        
        # Delete the temporary directory we used for detection images
        if not options.keep_rendered_frames:
            try:
                if options.force_rendered_frame_folder_deletion:
                    shutil.rmtree(rendering_output_dir)
                else:
                    for rendered_frame_fn in detected_frame_files:
                        os.remove(rendered_frame_fn)
            except Exception as e:
                print('Warning: error deleting rendered frames from folder {}:\n{}'.format(
                    rendering_output_dir,str(e)))
                pass
    
    # ...if we're rendering video
    
    
    ## (Optionally) delete the extracted frames
    
    if not options.keep_extracted_frames:
        
        try:
            if options.force_extracted_frame_folder_deletion:
                print('Recursively deleting frame output folder {}'.format(frame_output_folder))
                shutil.rmtree(frame_output_folder)
            else:
                for extracted_frame_fn in frame_filenames:
                    os.remove(extracted_frame_fn)
        except Exception as e:
            print('Warning: error removing extracted frames from folder {}:\n{}'.format(
                frame_output_folder,str(e)))
            pass
        
    return results

# ...process_video()


def process_video_folder(options):
    """
    Process a folder of videos through MD
    
    Args: 
        options (ProcessVideoOptions): all the parameters used to control this process,
            including filenames; see ProcessVideoOptions for details            
    """
    
    ## Validate options

    assert os.path.isdir(options.input_video_file), \
        '{} is not a folder'.format(options.input_video_file)
           
    assert options.output_json_file is not None, \
        'When processing a folder, you must specify an output .json file'
                         
    assert options.output_json_file.endswith('.json')
    video_json = options.output_json_file
    frames_json = options.output_json_file.replace('.json','.frames.json')
    os.makedirs(os.path.dirname(video_json),exist_ok=True)
    
    
    ## Split every video into frames
    
    if options.frame_folder is not None:
        frame_output_folder = options.frame_folder
    else:
        tempdir = os.path.join(tempfile.gettempdir(), 'process_camera_trap_video')
        os.makedirs(tempdir,exist_ok=True)
        
        # TODO: see above; this is a lazy fix to a permissions issue
        try:
            os.chmod(tempdir,0o777)
        except Exception:
            pass

        frame_output_folder = os.path.join(
            tempdir, os.path.basename(options.input_video_file) + '_frames_' + str(uuid1()))

    os.makedirs(frame_output_folder, exist_ok=True)

    print('Extracting frames')
    frame_filenames, Fs, video_filenames = \
        video_folder_to_frames(input_folder=options.input_video_file,
                               output_folder_base=frame_output_folder, 
                               recursive=options.recursive, 
                               overwrite=(not options.reuse_frames_if_available),
                               n_threads=options.n_cores,every_n_frames=options.frame_sample,
                               verbose=options.verbose)
    
    image_file_names = list(itertools.chain.from_iterable(frame_filenames))
    
    if len(image_file_names) == 0:
        if len(video_filenames) == 0:
            print('No videos found in folder {}'.format(options.input_video_file))
        else:
            print('No frames extracted from folder {}, this may be due to an '\
                  'unsupported video codec'.format(options.input_video_file))
        return

    if options.debug_max_frames is not None and options.debug_max_frames > 0:
        image_file_names = image_file_names[0:options.debug_max_frames]
    
    
    ## Run MegaDetector on the extracted frames
    
    if options.reuse_results_if_available and \
        os.path.isfile(frames_json):
            print('Loading results from {}'.format(frames_json))
            results = None
    else:
        print('Running MegaDetector')
        results = run_detector_batch.load_and_run_detector_batch(
            options.model_file, image_file_names,
            confidence_threshold=options.json_confidence_threshold,
            n_cores=options.n_cores,
            quiet=(not options.verbose),
            class_mapping_filename=options.class_mapping_filename)
    
        run_detector_batch.write_results_to_file(
            results, frames_json,
            relative_path_base=frame_output_folder,
            detector_file=options.model_file,
            custom_metadata={'video_frame_rate':Fs})
    
    
    ## Convert frame-level results to video-level results

    print('Converting frame-level results to video-level results')
    frame_results_to_video_results(frames_json,video_json)


    ## (Optionally) render output videos
    
    if options.render_output_video:

        # Render detections to images
        if options.frame_rendering_folder is not None:
            frame_rendering_output_dir = options.frame_rendering_folder
        else:
            frame_rendering_output_dir = os.path.join(
                tempdir, os.path.basename(options.input_video_file) + '_detections')
        
        # TODO: keep track of whether we created this folder, delete if we're deleting the rendered
        # frames and we created the folder, and the output files aren't in the same folder.  For now,
        # we're just deleting the rendered frames and leaving the empty folder around in this case.
        os.makedirs(frame_rendering_output_dir,exist_ok=True)
        
        detected_frame_files = visualize_detector_output.visualize_detector_output(
            detector_output_path=frames_json,
            out_dir=frame_rendering_output_dir,
            images_dir=frame_output_folder,
            confidence_threshold=options.rendering_confidence_threshold,
            preserve_path_structure=True,
            output_image_width=-1)
        
        # Choose an output folder
        output_folder_is_input_folder = False
        if options.output_video_file is not None:
            if os.path.isfile(options.output_video_file):
                raise ValueError('Rendering videos for a folder, but an existing file was specified as output')
            elif options.output_video_file == options.input_video_file:
                output_folder_is_input_folder = True
                output_video_folder = options.input_video_file
            else:
                os.makedirs(options.output_video_file,exist_ok=True)
                output_video_folder = options.output_video_file
        else:
            output_folder_is_input_folder = True
            output_video_folder = options.input_video_file
                                
        # For each video
        #
        # TODO: parallelize this loop
        #
        # i_video=0; input_video_file_abs = video_filenames[i_video]
        for i_video,input_video_file_abs in enumerate(video_filenames):
            
            video_fs = Fs[i_video]
            
            if options.frame_sample is None:                
                rendering_fs = video_fs
            else:
                rendering_fs = video_fs / options.frame_sample
            
            input_video_file_relative = os.path.relpath(input_video_file_abs,options.input_video_file)
            video_frame_output_folder = os.path.join(frame_rendering_output_dir,input_video_file_relative)
            assert os.path.isdir(video_frame_output_folder), \
                'Could not find frame folder for video {}'.format(input_video_file_relative)
            
            # Find the corresponding rendered frame folder
            video_frame_files = [fn for fn in detected_frame_files if \
                                 fn.startswith(video_frame_output_folder)]
            assert len(video_frame_files) > 0, 'Could not find rendered frames for video {}'.format(
                input_video_file_relative)
            
            from md_utils.path_utils import insert_before_extension
            
            # Select the output filename for the rendered video
            if output_folder_is_input_folder:
                video_output_file = insert_before_extension(input_video_file_abs,'annotated','_')
            else:
                video_output_file = os.path.join(output_video_folder,input_video_file_relative)
            
            os.makedirs(os.path.dirname(video_output_file),exist_ok=True)
            
            # Create the output video            
            print('Rendering detections for video {} to {} at {} fps (original video {} fps)'.format(
                input_video_file_relative,video_output_file,rendering_fs,video_fs))
            frames_to_video(video_frame_files, rendering_fs, video_output_file, codec_spec=options.fourcc)
                
        # ...for each video
        
        # Possibly clean up rendered frames
        if not options.keep_rendered_frames:
            try:
                if options.force_rendered_frame_folder_deletion:
                    shutil.rmtree(frame_rendering_output_dir)
                else:
                    for rendered_frame_fn in detected_frame_files:
                        os.remove(rendered_frame_fn)
            except Exception as e:
                print('Warning: error deleting rendered frames from folder {}:\n{}'.format(
                    frame_rendering_output_dir,str(e)))
                pass
        
    # ...if we're rendering video
    
    
    ## (Optionally) delete the extracted frames
    
    if not options.keep_extracted_frames:
        try:
            print('Deleting frame cache')
            if options.force_extracted_frame_folder_deletion:
                print('Recursively deleting frame output folder {}'.format(frame_output_folder))
                shutil.rmtree(frame_output_folder)
            else:
                for frame_fn in image_file_names:
                    os.remove(frame_fn)
        except Exception as e:
            print('Warning: error deleting frames from folder {}:\n{}'.format(
                frame_output_folder,str(e)))
            pass

# ...process_video_folder()


def options_to_command(options):
    
    cmd = 'python process_video.py'
    cmd += ' "' + options.model_file + '"'
    cmd += ' "' + options.input_video_file + '"'
    
    if options.recursive:
        cmd += ' --recursive'
    if options.frame_folder is not None:
        cmd += ' --frame_folder' + ' "' + options.frame_folder + '"'
    if options.frame_rendering_folder is not None:
        cmd += ' --frame_rendering_folder' + ' "' + options.frame_rendering_folder + '"'    
    if options.output_json_file is not None:
        cmd += ' --output_json_file' + ' "' + options.output_json_file + '"'
    if options.output_video_file is not None:
        cmd += ' --output_video_file' + ' "' + options.output_video_file + '"'
    if options.keep_extracted_frames:
        cmd += ' --keep_extracted_frames'
    if options.reuse_results_if_available:
        cmd += ' --reuse_results_if_available'    
    if options.reuse_frames_if_available:
        cmd += ' --reuse_frames_if_available'
    if options.render_output_video:
        cmd += ' --render_output_video'
    if options.keep_rendered_frames:
        cmd += ' --keep_rendered_frames'    
    if options.rendering_confidence_threshold is not None:
        cmd += ' --rendering_confidence_threshold ' + str(options.rendering_confidence_threshold)
    if options.json_confidence_threshold is not None:
        cmd += ' --json_confidence_threshold ' + str(options.json_confidence_threshold)
    if options.n_cores is not None:
        cmd += ' --n_cores ' + str(options.n_cores)
    if options.frame_sample is not None:
        cmd += ' --frame_sample ' + str(options.frame_sample)
    if options.debug_max_frames is not None:
        cmd += ' --debug_max_frames ' + str(options.debug_max_frames)
    if options.class_mapping_filename is not None:
        cmd += ' --class_mapping_filename ' + str(options.class_mapping_filename)
    if options.fourcc is not None:
        cmd += ' --fourcc ' + options.fourcc

    return cmd

    
#%% Interactive driver

if False:    
        
    #%% Process a folder of videos
    
    model_file = 'MDV5A'
    input_dir = r'c:\git\MegaDetector\test_images\test_images'    
    frame_folder = r'g:\temp\video_test\frames'
    rendering_folder = r'g:\temp\video_test\rendered-frames'
    output_json_file = r'g:\temp\video_test\video-test.json'
    output_video_folder = r'g:\temp\video_test\output_videos'    
    
    print('Processing folder {}'.format(input_dir))
    
    options = ProcessVideoOptions()    
    options.model_file = model_file
    options.input_video_file = input_dir
    options.output_video_file = output_video_folder
    options.frame_folder = frame_folder
    options.output_json_file = output_json_file
    options.frame_rendering_folder = rendering_folder
    options.render_output_video = True
    options.keep_extracted_frames = True
    options.keep_rendered_frames = True
    options.recursive = True
    options.reuse_frames_if_available = True
    options.reuse_results_if_available = True
    # options.confidence_threshold = 0.15
    # options.fourcc = 'mp4v'        
    
    cmd = options_to_command(options)
    print(cmd)
    # import clipboard; clipboard.copy(cmd)
    
    if False:
        process_video_folder(options)
        
    
    #%% Process a single video

    fn = os.path.expanduser('~/tmp/video-test/test-video.mp4')
    model_file = 'MDV5A'
    input_video_file = fn
    frame_folder = os.path.expanduser('~/tmp/video-test/frames')
    rendering_folder = os.path.expanduser('~/tmp/video-test/rendered-frames')
    
    options = ProcessVideoOptions()
    options.model_file = model_file
    options.input_video_file = input_video_file
    options.frame_folder = frame_folder
    options.frame_rendering_folder = rendering_folder
    options.render_output_video = True
    options.output_video_file = os.path.expanduser('~/tmp/video-test/detections.mp4')
    
    cmd = options_to_command(options)
    print(cmd)
    # import clipboard; clipboard.copy(cmd)
    
    if False:        
        process_video(options)    
            
    
#%% Command-line driver

def main():

    default_options = ProcessVideoOptions()
    
    parser = argparse.ArgumentParser(description=(
        'Run MegaDetector on each frame (or every Nth frame) in a video (or folder of videos), optionally '\
        'producing a new video with detections annotated'))

    parser.add_argument('model_file', type=str,
                        help='MegaDetector model file (.pt or .pb) or model name (e.g. "MDV5A")')

    parser.add_argument('input_video_file', type=str,
                        help='video file (or folder) to process')

    parser.add_argument('--recursive', action='store_true',
                        help='recurse into [input_video_file]; only meaningful if a folder '\
                         'is specified as input')
    
    parser.add_argument('--frame_folder', type=str, default=None,
                        help='folder to use for intermediate frame storage, defaults to a folder '\
                        'in the system temporary folder')
        
    parser.add_argument('--frame_rendering_folder', type=str, default=None,
                        help='folder to use for rendered frame storage, defaults to a folder in '\
                        'the system temporary folder')
    
    parser.add_argument('--output_json_file', type=str,
                        default=None, help='.json output file, defaults to [video file].json')

    parser.add_argument('--output_video_file', type=str,
                        default=None, help='video output file (or folder), defaults to '\
                            '[video file].mp4 for files, or [video file]_annotated for folders')

    parser.add_argument('--keep_extracted_frames',
                       action='store_true', help='Disable the deletion of extracted frames')
    
    parser.add_argument('--reuse_frames_if_available',
                       action='store_true', help="Don't extract frames that are already available in the frame extraction folder")
    
    parser.add_argument('--reuse_results_if_available',
                       action='store_true', help='If the output .json files exists, and this flag is set,'\
                           'we\'ll skip running MegaDetector')
    
    parser.add_argument('--render_output_video', action='store_true',
                        help='enable video output rendering (not rendered by default)')

    parser.add_argument('--fourcc', default=default_fourcc,
                        help='fourcc code to use for video encoding (default {}), only used if render_output_video is True'.format(default_fourcc))
    
    parser.add_argument('--keep_rendered_frames',
                       action='store_true', help='Disable the deletion of rendered (w/boxes) frames')

    parser.add_argument('--force_extracted_frame_folder_deletion',
                       action='store_true', help='By default, when keep_extracted_frames is False, we '\
                           'delete the frames, but leave the (probably-empty) folder in place.  This option '\
                           'forces deletion of the folder as well.  Use at your own risk; does not check '\
                           'whether other files were present in the folder.')
        
    parser.add_argument('--force_rendered_frame_folder_deletion',
                       action='store_true', help='By default, when keep_rendered_frames is False, we '\
                           'delete the frames, but leave the (probably-empty) folder in place.  This option '\
                           'forces deletion of the folder as well.  Use at your own risk; does not check '\
                           'whether other files were present in the folder.')
        
    parser.add_argument('--rendering_confidence_threshold', type=float,
                        default=None, help="don't render boxes with confidence below this threshold (defaults to choosing based on the MD version)")

    parser.add_argument('--json_confidence_threshold', type=float,
                        default=0.0, help="don't include boxes in the .json file with confidence "\
                            'below this threshold (default {})'.format(
                                default_options.json_confidence_threshold))

    parser.add_argument('--n_cores', type=int,
                        default=1, help='number of cores to use for frame separation and detection. '\
                            'If using a GPU, this option will be respected for frame separation but '\
                            'ignored for detection.  Only relevant to frame separation when processing '\
                            'a folder.')

    parser.add_argument('--frame_sample', type=int,
                        default=None, help='process every Nth frame (defaults to every frame)')

    parser.add_argument('--debug_max_frames', type=int,
                        default=-1, help='trim to N frames for debugging (impacts model execution, '\
                            'not frame rendering)')
    
    parser.add_argument('--class_mapping_filename',
                        type=str,
                        default=None, help='Use a non-default class mapping, supplied in a .json file '\
                            'with a dictionary mapping int-strings to strings.  This will also disable '\
                            'the addition of "1" to all category IDs, so your class mapping should start '\
                            'at zero.')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()
        
    args = parser.parse_args()
    options = ProcessVideoOptions()
    args_to_object(args,options)

    if os.path.isdir(options.input_video_file):
        process_video_folder(options)
    else:
        process_video(options)

if __name__ == '__main__':
    main()



"""

taxonomy_csv_checker.py

Checks the taxonomy CSV file to make sure that for each row:

1) The 'taxonomy_level' column matches the lowest-level taxon level in the
    'taxonomy_string' column.

2) The 'scientific_name' column matches the scientific name from the
    lowest-level taxon level in the 'taxonomy_string' column.

Prints out any mismatches.

Also prints out nodes that have 2 ambiguous parents. See "CASE 2" from the
module docstring of taxonomy_graph.py.

"""

#%% Imports

import sys
import argparse

import networkx as nx
import pandas as pd

from typing import Optional

from taxonomy_mapping.taxonomy_graph import TaxonNode, dag_to_tree


#%% Taxonomy checking

def check_taxonomy_csv(csv_path: str) -> None:
    """
    See module docstring.
    """
    
    taxonomy_df = pd.read_csv(csv_path)

    graph = nx.DiGraph()
    taxon_to_node = {}  # maps (taxon_level, taxon_name) to a TaxonNode

    num_taxon_level_errors = 0
    num_scientific_name_errors = 0

    for i_row, row in taxonomy_df.iterrows():
        
        ds = row['dataset_name']
        ds_label = row['query']
        scientific_name = row['scientific_name']
        level = row['taxonomy_level']
        
        # This used to represent the source of the mapping: iNat, gbif, or manual.  We've
        # stopped tracking this, so this is now vestigial.
        id_source = 0 # row['source']

        taxa_ancestry = row['taxonomy_string']
        if pd.isna(taxa_ancestry):
            # taxonomy CSV rows without 'taxonomy_string' entries are excluded
            # from the taxonomy graph, but can be included in a classification
            # label specification JSON via the 'dataset_labels' key
            continue
        else:
            taxa_ancestry = eval(taxa_ancestry)  # pylint: disable=eval-used

        taxon_child: Optional[TaxonNode] = None
        for j, taxon in enumerate(taxa_ancestry):
            taxon_id, taxon_level, taxon_name, _ = taxon

            key = (taxon_level, taxon_name)
            if key not in taxon_to_node:
                taxon_to_node[key] = TaxonNode(level=taxon_level,
                                               name=taxon_name, graph=graph)
            node = taxon_to_node[key]

            if taxon_child is not None:
                node.add_child(taxon_child)

            node.add_id(id_source, int(taxon_id))  # np.int64 -> int
            if j == 0:
                if level != taxon_level:
                    print(f'row: {i_row}, {ds}, {ds_label}')
                    print(f'- taxonomy_level column: {level}, '
                          f'level from taxonomy_string: {taxon_level}')
                    print()
                    num_taxon_level_errors += 1

                if scientific_name != taxon_name:
                    print(f'row: {i_row}, {ds}, {ds_label}')
                    print(f'- scientific_name column: {scientific_name}, '
                          f'name from taxonomy_string: {taxon_name}')
                    print()
                    num_scientific_name_errors += 1

            taxon_child = node
    
    # ...for each row in the taxonomy file        

    assert nx.is_directed_acyclic_graph(graph)

    for node in graph.nodes:
        assert len(node.parents) <= 2
        if len(node.parents) == 2:
            p0 = node.parents[0]
            p1 = node.parents[1]
            assert p0 is not p1

            p0_is_ancestor_of_p1 = p1 in nx.descendants(graph, p0)
            p1_is_ancestor_of_p0 = p0 in nx.descendants(graph, p1)
            if not p0_is_ancestor_of_p1 and not p1_is_ancestor_of_p0:
                print('Node with two ambiguous parents:', node)
                print('\t', p0)
                print('\t\t', p0.parents)
                print('\t', p1)
                print('\t\t', p1.parents)

    try:
        dag_to_tree(graph, taxon_to_node)
        print('All ambiguous parents have hard-coded resolution in '
              'dag_to_tree().')
    except AssertionError as e:
        print(f'At least one node has unresolved ambiguous parents: {e}')

    print('Processed {} rows from {}'.format(len(taxonomy_df),csv_path))
    
    print('num taxon level errors:', num_taxon_level_errors)
    print('num scientific name errors:', num_scientific_name_errors)


#%% Command-line driver
    
if __name__ == '__main__':
    
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'taxonomy_csv_path',
        help='path to taxonomy CSV file')
    
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()
    
    args = parser.parse_args()

    check_taxonomy_csv(args.taxonomy_csv_path)


#%% Interactive driver
    
if False:
    
    #%%
    
    import os
    csv_path = os.path.expanduser('~/lila/lila-taxonomy-mapping_release.csv')
    check_taxonomy_csv(csv_path)
    



#
# xmp_integration.py
#
# Tools for loading MegaDetector batch API results into XMP metadata, specifically
# for consumption in digiKam:
#
# https://cran.r-project.org/web/packages/camtrapR/vignettes/camtrapr2.html
#
    
#%% Imports and constants

import argparse
import tkinter
from tkinter import ttk, messagebox, filedialog

import inspect
import os
import sys
import json
import pyexiv2
import ntpath
import threading
import traceback

from tqdm import tqdm
from multiprocessing import Pool
from multiprocessing.pool import ThreadPool
from functools import partial

category_mapping = {'person': 'Human', 'animal': 'Animal', 'vehicle': 'Vehicle'}


#%% Class definitions

class xmp_gui:
    
    root = None
    textarea_min_threshold = None
    textarea_status = None
    textarea_remove_path = None
    textarea_rename_conf = None
    textarea_rename_cats = None
    num_threads = 1
    
class xmp_integration_options:
    
    # Folder where images are stored
    image_folder = None
    
    # .json file containing MegaDetector output
    input_file = None
    
    # String to remove from all path names, typically representing a 
    # prefix that was added during MegaDetector processing
    remove_path = None
    
    # Optionally *rename* (not copy) all images that have no detections
    # above [rename_conf] for the categories in rename_cats from x.jpg to
    # x.check.jpg
    rename_conf = None
    
    # Comma-deleted list of category names (or "all") to apply the rename_conf
    # behavior to.
    rename_cats = None
    
    # Minimum detection threshold (applies to all classes, defaults to None,
    # i.e. 0.0
    min_threshold = None
    num_threads = 1
    xmp_gui = None
    
    
#%% Functions
  
def write_status(options,s):
    
    if options.xmp_gui is None:
        return
    options.xmp_gui.textarea_status.configure(state="normal")
    options.xmp_gui.textarea_status.insert(tkinter.END, s + '\n')
    options.xmp_gui.textarea_status.configure(state="disabled")

    
n_images_processed = 0
    
def update_xmp_metadata(categories, options, rename_cats, n_images, image):
    """
    Update the XMP metadata for a single image
    """
    
    # Relative image path
    filename = ''
    
    # Absolute image path
    img_path = ''
    
    global n_images_processed
    
    try:
        
        filename = image['file']
        if options.remove_path != None and len(options.remove_path) > 0:
            filename = filename.replace(options.remove_path, '')
        img_path = os.path.join(options.image_folder, filename)
        assert os.path.isfile(img_path), 'Image {} not found'.format(img_path)
        
        # List of categories to write to XMP metadata
        image_categories = []
        
        # Categories with above-threshold detections present for
        # this image
        original_image_cats = []
        
        # Maximum confidence for each category
        original_image_cats_conf = {}
        
        for detection in image['detections']:
            
            cat = category_mapping[categories[detection['category']]]
            
            # Have we already added this to the list of categories to
            # write out to this image?
            if cat not in image_categories:
                
                # If we're supposed to compare to a threshold...
                if len(options.min_threshold) > 0 and \
                    options.min_threshold != None:
                    if float(detection['conf']) > float(options.min_threshold):
                        image_categories.append(cat)
                        original_image_cats.append(
                            categories[detection['category']])
                        
                # Else we treat *any* detection as valid...
                else:
                    image_categories.append(cat)
                    original_image_cats.append(categories[detection['category']])

            # Keep track of the highest-confidence detection for this class                
            if options.min_threshold != None and \
                len(options.min_threshold) > 0 and \
                    detection['conf'] > \
                        original_image_cats_conf.get(
                            categories[detection['category']], 0):
                            
                original_image_cats_conf[categories[detection['category']]] = \
                    detection['conf']
                    
        img = pyexiv2.Image(r'{0}'.format(img_path))
        img.modify_xmp({'Xmp.lr.hierarchicalSubject': image_categories})
        
        # If we're doing the rename/.check behavior...
        if not (options.rename_conf is None and options.rename_cats is None):
            
            matching_cats = set(rename_cats).intersection(set(original_image_cats))
            is_conf_low = False
            if options.min_threshold != None and len(options.min_threshold) > 0:
                for matching_cat in matching_cats:
                    if original_image_cats_conf[matching_cat] < float(options.rename_conf):
                        is_conf_low = True
            if options.min_threshold != None and \
                len(options.min_threshold) > 0 and \
                    len(image['detections']) == 0 or \
                (len(options.rename_conf) > 0 and \
                is_conf_low is True and \
                    len(matching_cats) > 0):
                        
                parent_folder = os.path.dirname(img_path)
                file_name = ntpath.basename(img_path)
                manual_file_name = file_name.split('.')[0]+'_check' + '.' + file_name.split('.')[1]
                os.rename(img_path, os.path.join(parent_folder, manual_file_name))
                
        if options.xmp_gui is not None:
            
            n_images_processed += 1
            percentage = round((n_images_processed)/n_images*100)
            options.xmp_gui.progress_bar['value'] = percentage
            options.xmp_gui.root.update_idletasks()
            options.xmp_gui.style.configure('text.Horizontal.Tprogress_bar',
                            text='{:g} %'.format(percentage))
                
    except Exception as e:
    
        s = 'Error processing image {}: {}'.format(filename,str(e))
        print(s)
        traceback.print_exc()
        write_status(options,s)
        
        if False:
            
            # Legacy code to rename files where XMP writing failed
            parent_folder = os.path.dirname(img_path)
            file_name = ntpath.basename(img_path)        
            failed_file_name = file_name.split('.')[0]+'_failed' + '.' + file_name.split('.')[1]
            os.rename(img_path, os.path.join(
                parent_folder, failed_file_name))


def process_input_data(options):
    """
    Main function to loop over images and modify XMP data
    """
    
    if options.xmp_gui is not None:
        
        if (options.image_folder is None) or (len(options.image_folder) == 0):
            tkinter.messagebox.showerror(title='Error', message='Image folder is not selected')
            sys.exit()
        if (options.input_file is None) or (len(options.input_file) == 0):
            tkinter.messagebox.showerror(
                title='Error', message='No MegaDetector .json file selected')
            sys.exit()
        options.remove_path = options.xmp_gui.textarea_remove_path.get()
        options.rename_conf = options.xmp_gui.textarea_rename_conf.get()
        options.rename_cats = options.xmp_gui.textarea_rename_cats.get()
        options.num_threads = options.xmp_gui.textarea_num_threads.get()
        options.min_threshold = options.xmp_gui.textarea_min_threshold.get()
            
    try:
        
        with open(options.input_file, 'r') as f:
            data = f.read()
    
        data = json.loads(data)
        categories = data['detection_categories']
    
        images = data['images']
        n_images = len(images)
        if not (options.rename_conf is None and options.rename_cats is None):
            rename_cats = options.rename_cats.split(",")
            if rename_cats[0] == 'all':
                rename_cats = list(category_mapping.keys())
        else:
            rename_cats = []
        if len(options.num_threads) > 0:
            num_threads = int(options.num_threads)
        else:
            num_threads = 1
        print(num_threads)
        if options.xmp_gui is None:
            func = partial(update_xmp_metadata, categories, options, rename_cats, n_images)
            with Pool(num_threads) as p:
                with tqdm(total=n_images) as pbar:
                    for i, _ in enumerate(p.imap_unordered(func, images)):
                        pbar.update()
        else:
            func = partial(update_xmp_metadata, categories, options, rename_cats, n_images)
            with ThreadPool(num_threads) as p:
                p.map(func, images)
            s = 'Successfully processed {} images'.format(n_images)
            print(s)
            write_status(options,s)
        
    except Exception as e:
        
        print('Error processing input data: {}'.format(str(e)))
        traceback.print_exc()
        if options.xmp_gui is not None:
            tkinter.messagebox.showerror(title='Error', 
                                         message='Make Sure you selected the proper image folder and JSON files')
        sys.exit()


def start_input_processing(options):
    
    t = threading.Thread(target=lambda: process_input_data(options))
    t.start()


def browse_folder(options,folder_path_var):
    
    filename = tkinter.filedialog.askdirectory()
    options.image_folder = r'{0}'.format(filename)
    folder_path_var.set(filename)


def browse_file(options,file_path_var):
    
    filename = tkinter.filedialog.askopenfilename()
    options.input_file = r'{0}'.format(filename)
    file_path_var.set(filename)


def create_gui(options):
    
    root = tkinter.Tk()
    root.resizable(False, False)
    root.configure(background='white')
    root.title('DigiKam Integration')
    
    group = tkinter.LabelFrame(root, padx=5, pady=5)
    group.configure(background = 'white')
    group.pack(padx=10, pady=10, fill='both', expand='yes')

    canvas = tkinter.Canvas(group, width = 800, height = 150)    
    canvas.configure(background = 'white')
    canvas.pack()      
    img1 = tkinter.PhotoImage(file='images/aiforearth.png')      
    canvas.create_image(0,0, anchor=tkinter.NW, image=img1)
    img2 = tkinter.PhotoImage(file='images/bg.png')      
    canvas.create_image(0,20, anchor=tkinter.NW, image=img2)

    frame = tkinter.Frame(root)
    frame.configure(background='white')
    frame.pack()
    
    l1 = tkinter.Label(frame, text='Folder containing images')
    l1.configure(background='white')
    l1.grid(row=0, column=0)
    
    folder_path_var = tkinter.StringVar()
    
    e1 = tkinter.Entry(frame, width=50, textvariable=folder_path_var, highlightthickness=1) 
    e1.configure(highlightbackground='grey', highlightcolor='grey')
    e1.grid(row=0, column=2)
    
    b1 = tkinter.Button(frame, text='Browse', fg='blue', command=lambda: browse_folder(options,folder_path_var))
    b1.grid(row=0, column=5, padx=10)
    
    l2 = tkinter.Label(frame, text='Path to MegaDetector output .json file') 
    l2.configure(background='white')
    l2.grid(row=1, column=0)
    
    file_path_var = tkinter.StringVar()
    
    e2 = tkinter.Entry(frame, width=50, textvariable=file_path_var, highlightthickness=1)
    e2.configure(highlightbackground='grey', highlightcolor='grey')    
    e2.grid(row=1, column=2)
    
    b2 = tkinter.Button(frame, text='Browse', fg='blue', command=lambda: browse_file(options,file_path_var))
    b2.grid(row=1, column=5, padx=10)

    l6 = tkinter.Label(frame, text='Minimum confidence to consider a category') 
    l6.configure(background='white')
    l6.grid(row=2, column=0)
    
    textarea_min_threshold = tkinter.Entry(frame, width=50, highlightthickness=1)
    textarea_min_threshold.configure(highlightbackground='grey', highlightcolor='grey')
    textarea_min_threshold.grid(row=2, column=2)
    
    l3 = tkinter.Label(frame, text='Prefix to remove from image paths (optional)') 
    l3.configure(background='white')
    l3.grid(row=3, column=0)
    
    textarea_remove_path = tkinter.Entry(frame, width=50, highlightthickness=1)
    textarea_remove_path.configure(highlightbackground='grey', highlightcolor='grey')
    textarea_remove_path.grid(row=3, column=2)

    l4 = tkinter.Label(frame, text='Confidence level to move images requires manual check (optional)') 
    l4.configure(background='white')
    l4.grid(row=4, column=0)

    textarea_rename_conf = tkinter.Entry(frame, width=50, highlightthickness=1)
    textarea_rename_conf.configure(highlightbackground='grey', highlightcolor='grey')
    textarea_rename_conf.grid(row=4, column=2)


    l5 = tkinter.Label(frame, text='Categories to check for the confidence (optional)') 
    l5.configure(background='white')
    l5.grid(row=5, column=0)

    textarea_rename_cats = tkinter.Entry(frame, width=50, highlightthickness=1)
    textarea_rename_cats.configure(highlightbackground='grey', highlightcolor='grey')
    textarea_rename_cats.grid(row=5, column=2)

    l6 = tkinter.Label(frame, text='Number of threads to run (optional)') 
    l6.configure(background='white')
    l6.grid(row=6, column=0)

    textarea_num_threads = tkinter.Entry(frame, width=50, highlightthickness=1)
    textarea_num_threads.configure(highlightbackground='grey', highlightcolor='grey')
    textarea_num_threads.grid(row=6, column=2)
    
    sb = tkinter.Button(frame, text='Submit', fg='black',
                command=lambda: start_input_processing(options), padx=10)
    sb.grid(row=7, column=2, padx=10, pady=10)

    style = tkinter.ttk.Style(root)
    style.layout('text.Horizontal.Tprogress_bar',
                [('Horizontal.progress_bar.trough',
                {'children': [('Horizontal.progress_bar.pbar',
                                {'side': 'left', 'sticky': 'ns'})],
                    'sticky': 'nswe'}),
                ('Horizontal.progress_bar.label', {'sticky': ''})])    
    style.configure('text.Horizontal.Tprogress_bar', text='0 %')

    progress_bar = tkinter.ttk.Progressbar(root, style='text.Horizontal.Tprogress_bar', length=700,
                                maximum=100, value=0, mode='determinate')
    progress_bar.pack(pady=10)

    group2 = tkinter.LabelFrame(root, text='Status', padx=5, pady=5)
    group2.pack(padx=10, pady=10, fill='both', expand='yes')

    textarea_status = tkinter.Text(group2, height=10, width=100)
    textarea_status.configure(state="disabled")
    textarea_status.pack()
    
    options.xmp_gui = xmp_gui()
    options.xmp_gui.root = root
    options.xmp_gui.textarea_min_threshold = textarea_min_threshold
    options.xmp_gui.textarea_remove_path = textarea_remove_path
    options.xmp_gui.textarea_rename_conf = textarea_rename_conf
    options.xmp_gui.textarea_rename_cats = textarea_rename_cats
    options.xmp_gui.textarea_num_threads = textarea_num_threads
    options.xmp_gui.textarea_status = textarea_status
    options.xmp_gui.progress_bar = progress_bar
    options.xmp_gui.style = style
    
    root.mainloop()
    
    
#%% Interactive/test driver

if False:
    
    #%%
    
    options = xmp_integration_options()
    options.input_file = r"C:\temp\demo_images\ssmini_xmp_test_orig\ssmini.mdv4.json"
    options.image_folder = r"C:\temp\demo_images\ssmini_xmp_test"
    options.remove_path = 'my_images/'
    process_input_data(options)
    
    
#%% Command-line driver

def args_to_object(args,obj):
    """
    Copy all fields from the argparse table "args" to the object "obj"
    """
    for n, v in inspect.getmembers(args):
        if not n.startswith('_'):
            setattr(obj, n, v)


def main():
    
    parser = argparse.ArgumentParser()
    parser.add_argument('--input_file', help = 'Path to the MegaDetector .json file', default=None)
    parser.add_argument('--image_folder', help = 'Path to the folder containing images', default=None)
    parser.add_argument('--min_threshold', help = 'Minimum detection confidence that will be treated as a detection event', default=None)
    parser.add_argument('--remove_path', help = 'Prefix to remove from image paths in the .json file', default=None)
    parser.add_argument('--rename_conf', help = 'Below this confidence level, images will be renamed for manual check', default=None)
    parser.add_argument('--rename_cat', help = 'Category (or comma-delimited categories) to apply renaming behavior to', default=None)
    parser.add_argument('--num_threads', help = 'Number of threads to use for image processing', default=1)
    parser.add_argument('--gui', help = 'Run in GUI mode', action='store_true')
    
    options = xmp_integration_options()
    args = parser.parse_args()
    args_to_object(args,options)
    
    if options.gui:
        assert options.input_file is None, 'Command-line argument specified in GUI mode'
        assert options.image_folder is None, 'Command-line argument specified in GUI mode'
        assert options.min_threshold is None, 'Command-line argument specified in GUI mode'
        assert options.remove_path is None, 'Command-line argument specified in GUI mode'
        assert options.rename_conf is None, 'Command-line argument specified in GUI mode'
        assert options.rename_cat is None, 'Command-line argument specified in GUI mode'
        assert options.num_threads == 1, 'Command-line argument specified in GUI mode'
        create_gui(options)    
    else:
        process_input_data(options)

if __name__ == '__main__':
    
    main()



"""

json_to_azcopy_list.py

Given a queried_images.json file output from json_validator.py, generates
one text file <dataset>_images.txt for every dataset included.

See: https://github.com/Azure/azure-storage-azcopy/wiki/Listing-specific-files-to-transfer

"""

#%% Imports and constants 

import json
import os

from tqdm import tqdm

from data_management.megadb import megadb_utils
from md_utils import sas_blob_utils

images_dir = ''
queried_images_json_path = 'run_idfg2/queried_images.json'
output_dir = 'run_idfg2/'


#%% Prepare AzCopy list

with open(queried_images_json_path, 'r') as f:
    js = json.load(f)

datasets_table = megadb_utils.MegadbUtils().get_datasets_table()

output_files = {}

pbar = tqdm(js.items())
for img_path, img_info in pbar:
    save_path = os.path.join(images_dir, img_path)
    if os.path.exists(save_path):
        continue

    ds, img_file = img_path.split('/', maxsplit=1)
    if ds not in output_files:
        output_path = os.path.join(output_dir, f'{ds}_images.txt')
        output_files[ds] = open(output_path, 'w')

        dataset_info = datasets_table[ds]
        account = dataset_info['storage_account']
        container = dataset_info['container']

        if 'public' in datasets_table[ds]['access']:
            url = sas_blob_utils.build_azure_storage_uri(
                account, container)
        else:
            url = sas_blob_utils.build_azure_storage_uri(
                account, container,
                sas_token=dataset_info['container_sas_key'][1:])
        pbar.write(f'"{url}"')

    output_files[ds].write(img_file + '\n')

for f in output_files.values():
    f.close()



########
#
# run_detector_batch.py
#
# Module to run MegaDetector on lots of images, writing the results
# to a file in the MegaDetector output format:
# 
# https://github.com/agentmorris/MegaDetector/tree/main/api/batch_processing#megadetector-batch-output-format
# 
# This enables the results to be used in our post-processing pipeline; see
# api/batch_processing/postprocessing/postprocess_batch_results.py .
# 
# This script can save results to checkpoints intermittently, in case disaster
# strikes. To enable this, set --checkpoint_frequency to n > 0, and results 
# will be saved as a checkpoint every n images. Checkpoints will be written 
# to a file in the same directory as the output_file, and after all images
# are processed and final results file written to output_file, the temporary
# checkpoint file will be deleted. If you want to resume from a checkpoint, set
# the checkpoint file's path using --resume_from_checkpoint.
# 
# The `threshold` you can provide as an argument is the confidence threshold above
# which detections will be included in the output file.
# 
# Has preliminary multiprocessing support for CPUs only; if a GPU is available, it will
# use the GPU instead of CPUs, and the --ncores option will be ignored.  Checkpointing
# is not supported when using a GPU.
# 
# Does not have a command-line option to bind the process to a particular GPU, but you can 
# prepend with "CUDA_VISIBLE_DEVICES=0 ", for example, to bind to GPU 0, e.g.:
# 
# CUDA_VISIBLE_DEVICES=0 python detection/run_detector_batch.py md_v4.1.0.pb ~/data ~/mdv4test.json 
#
# You can disable GPU processing entirely by setting CUDA_VISIBLE_DEVICES=''.
#
########

#%% Constants, imports, environment

import argparse
import json
import os
import sys
import time
import copy
import shutil
import warnings
import itertools
import humanfriendly

from datetime import datetime
from functools import partial
from tqdm import tqdm

import multiprocessing
from threading import Thread
from multiprocessing import Process, Manager

# Multiprocessing uses processes, not threads... leaving this here (and commented out)
# to make sure I don't change this casually at some point, it changes a number of 
# assumptions about interaction with PyTorch and TF.
# from multiprocessing.pool import ThreadPool as workerpool
from multiprocessing.pool import Pool as workerpool

import detection.run_detector as run_detector
from detection.run_detector import is_gpu_available,\
    load_detector,\
    try_download_known_detector,\
    get_detector_version_from_filename,\
    get_detector_metadata_from_version_string

from md_utils import path_utils
import md_visualization.visualization_utils as vis_utils
from data_management import read_exif

# Numpy FutureWarnings from tensorflow import
warnings.filterwarnings('ignore', category=FutureWarning)

# Number of images to pre-fetch
max_queue_size = 10

# How often should we print progress when using the image queue?
n_queue_print = 1000

use_threads_for_queue = False
verbose = False

exif_options = read_exif.ReadExifOptions()
exif_options.processing_library = 'pil'
exif_options.byte_handling = 'convert_to_string'


#%% Support functions for multiprocessing

def producer_func(q,image_files):
    """ 
    Producer function; only used when using the (optional) image queue.
    
    Reads up to N images from disk and puts them on the blocking queue for processing.
    """
    
    if verbose:
        print('Producer starting'); sys.stdout.flush()
        
    for im_file in image_files:
    
        try:
            if verbose:
                print('Loading image {}'.format(im_file)); sys.stdout.flush()
            image = vis_utils.load_image(im_file)
        except Exception:
            print('Producer process: image {} cannot be loaded.'.format(im_file))
            image = run_detector.FAILURE_IMAGE_OPEN            
        
        if verbose:
            print('Queueing image {}'.format(im_file)); sys.stdout.flush()
        q.put([im_file,image])
    
    q.put(None)
        
    print('Finished image loading'); sys.stdout.flush()
    
    
def consumer_func(q,return_queue,model_file,confidence_threshold,image_size=None):
    """ 
    Consumer function; only used when using the (optional) image queue.
    
    Pulls images from a blocking queue and processes them.
    """
    
    if verbose:
        print('Consumer starting'); sys.stdout.flush()

    start_time = time.time()
    detector = load_detector(model_file)
    elapsed = time.time() - start_time
    print('Loaded model (before queueing) in {}, printing updates every {} images'.format(
        humanfriendly.format_timespan(elapsed),n_queue_print))
    sys.stdout.flush()
        
    results = []
    
    n_images_processed = 0
    
    while True:
        r = q.get()
        if r is None:
            q.task_done()
            return_queue.put(results)
            return
        n_images_processed += 1
        im_file = r[0]
        image = r[1]
        if verbose or ((n_images_processed % n_queue_print) == 1):
            elapsed = time.time() - start_time
            images_per_second = n_images_processed / elapsed
            print('De-queued image {} ({:.2f}/s) ({})'.format(n_images_processed,
                                                          images_per_second,
                                                          im_file));
            sys.stdout.flush()
        if isinstance(image,str):
            # This is how the producer function communicates read errors
            results.append({'file': im_file,
                            'failure': image})
        else:
            results.append(process_image(im_file=im_file,detector=detector,
                                         confidence_threshold=confidence_threshold,
                                         image=image,quiet=True,image_size=image_size))
        if verbose:
            print('Processed image {}'.format(im_file)); sys.stdout.flush()
        q.task_done()
            

def run_detector_with_image_queue(image_files,model_file,confidence_threshold,
                                  quiet=False,image_size=None):
    """
    Driver function for the (optional) multiprocessing-based image queue; only used 
    when --use_image_queue is specified.  Starts a reader process to read images from disk, but 
    processes images in the  process from which this function is called (i.e., does not currently
    spawn a separate consumer process).
    """
    
    q = multiprocessing.JoinableQueue(max_queue_size)
    return_queue = multiprocessing.Queue(1)
    
    if use_threads_for_queue:
        producer = Thread(target=producer_func,args=(q,image_files,))
    else:
        producer = Process(target=producer_func,args=(q,image_files,))
    producer.daemon = False
    producer.start()
 
    # The queue system is a little more elegant if we start one thread for reading and one
    # for processing, and this works fine on Windows, but because we import TF at module load,
    # CUDA will only work in the main process, so currently the consumer function runs here.
    #
    # To enable proper multi-GPU support, we may need to move the TF import to a separate module
    # that isn't loaded until very close to where inference actually happens.
    run_separate_consumer_process = False

    if run_separate_consumer_process:
        if use_threads_for_queue:
            consumer = Thread(target=consumer_func,args=(q,return_queue,model_file,
                                                         confidence_threshold,image_size,))
        else:
            consumer = Process(target=consumer_func,args=(q,return_queue,model_file,
                                                          confidence_threshold,image_size,))
        consumer.daemon = True
        consumer.start()
    else:
        consumer_func(q,return_queue,model_file,confidence_threshold,image_size)

    producer.join()
    print('Producer finished')
   
    if run_separate_consumer_process:
        consumer.join()
        print('Consumer finished')
    
    q.join()
    print('Queue joined')

    results = return_queue.get()
    
    return results


#%% Other support functions

def chunks_by_number_of_chunks(ls, n):
    """
    Splits a list into n even chunks.

    Args
    - ls: list
    - n: int, # of chunks
    """
    
    for i in range(0, n):
        yield ls[i::n]


#%% Image processing functions

def process_images(im_files, detector, confidence_threshold, use_image_queue=False, 
                   quiet=False, image_size=None, checkpoint_queue=None, include_image_size=False,
                   include_image_timestamp=False, include_exif_data=False):
    """
    Runs MegaDetector over a list of image files.

    Args
    - im_files: list of str, paths to image files
    - detector: loaded model or str (path to .pb/.pt model file)
    - confidence_threshold: float, only detections above this threshold are returned

    Returns
    - results: list of dict, each dict represents detections on one image
        see the 'images' key in https://github.com/agentmorris/MegaDetector/tree/master/api/batch_processing#batch-processing-api-output-format
    """
    
    if isinstance(detector, str):
        start_time = time.time()
        detector = load_detector(detector)
        elapsed = time.time() - start_time
        print('Loaded model (batch level) in {}'.format(humanfriendly.format_timespan(elapsed)))

    if use_image_queue:
        run_detector_with_image_queue(im_files, detector, confidence_threshold, 
                                      quiet=quiet, image_size=image_size,
                                      include_image_size=include_image_size, 
                                      include_image_timestamp=include_image_timestamp,
                                      include_exif_data=include_exif_data)
    else:
        results = []
        for im_file in im_files:
            result = process_image(im_file, detector, confidence_threshold,
                                         quiet=quiet, image_size=image_size, 
                                         include_image_size=include_image_size, 
                                         include_image_timestamp=include_image_timestamp,
                                         include_exif_data=include_exif_data)

            if checkpoint_queue is not None:
                checkpoint_queue.put(result)
            results.append(result)                                    
            
        return results

# ...def process_images(...)


def process_image(im_file, detector, confidence_threshold, image=None, 
                  quiet=False, image_size=None, include_image_size=False,
                  include_image_timestamp=False, include_exif_data=False,
                  skip_image_resizing=False):
    """
    Runs MegaDetector on a single image file.

    Args
    - im_file: str, path to image file
    - detector: loaded model
    - confidence_threshold: float, only detections above this threshold are returned
    - image: previously-loaded image, if available
    - skip_image_resizing: whether to skip internal image resizing and rely on external resizing

    Returns:
    - result: dict representing detections on one image
        see the 'images' key in 
        https://github.com/agentmorris/MegaDetector/tree/master/api/batch_processing#batch-processing-api-output-format
    """
    
    if not quiet:
        print('Processing image {}'.format(im_file))
    
    if image is None:
        try:
            image = vis_utils.load_image(im_file)
        except Exception as e:
            if not quiet:
                print('Image {} cannot be loaded. Exception: {}'.format(im_file, e))
            result = {
                'file': im_file,
                'failure': run_detector.FAILURE_IMAGE_OPEN
            }
            return result

    try:
        result = detector.generate_detections_one_image(
            image, im_file, detection_threshold=confidence_threshold, image_size=image_size,
            skip_image_resizing=skip_image_resizing)
    except Exception as e:
        if not quiet:
            print('Image {} cannot be processed. Exception: {}'.format(im_file, e))
        result = {
            'file': im_file,
            'failure': run_detector.FAILURE_INFER
        }
        return result

    if include_image_size:
        result['width'] = image.width
        result['height'] = image.height

    if include_image_timestamp:
        result['datetime'] = get_image_datetime(image)

    if include_exif_data:
        result['exif_metadata'] = read_exif.read_pil_exif(image,exif_options)

    return result

# ...def process_image(...)


#%% Main function

def load_and_run_detector_batch(model_file, image_file_names, checkpoint_path=None,
                                confidence_threshold=run_detector.DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD,
                                checkpoint_frequency=-1, results=None, n_cores=1,
                                use_image_queue=False, quiet=False, image_size=None, class_mapping_filename=None, 
                                include_image_size=False, include_image_timestamp=False, 
                                include_exif_data=False):
    """
    Args
    - model_file: str,quiet path to .pb model file
    - image_file_names: list of strings (image filenames), a single image filename, 
                        a folder to recursively search for images in, or a .json file containing
                        a list of images.
    - checkpoint_path: str, path to JSON checkpoint file
    - confidence_threshold: float, only detections above this threshold are returned
    - checkpoint_frequency: int, write results to JSON checkpoint file every N images
    - results: list of dict, existing results loaded from checkpoint
    - n_cores: int, # of CPU cores to use
    - class_mapping_filename: str, use a non-default class mapping supplied in a .json file

    Returns
    - results: list of dicts; each dict represents detections on one image
    """
    
    if n_cores is None:
        n_cores = 1
    
    if confidence_threshold is None:
        confidence_threshold=run_detector.DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD
        
    if checkpoint_frequency is None:
        checkpoint_frequency = -1

    # This is an experimental hack to allow the use of non-MD YOLOv5 models through
    # the same infrastructure; it disables the code that enforces MDv5-like class lists.
    if class_mapping_filename is not None:
        run_detector.USE_MODEL_NATIVE_CLASSES = True
        with open(class_mapping_filename,'r') as f:
            class_mapping = json.load(f)
        print('Loaded custom class mapping:')
        print(class_mapping)
        run_detector.DEFAULT_DETECTOR_LABEL_MAP = class_mapping
        
    # Handle the case where image_file_names is not yet actually a list
    if isinstance(image_file_names,str):
        
        # Find the images to score; images can be a directory, may need to recurse
        if os.path.isdir(image_file_names):
            image_dir = image_file_names
            image_file_names = path_utils.find_images(image_dir, True)
            print('{} image files found in folder {}'.format(len(image_file_names),image_dir))
            
        # A json list of image paths
        elif os.path.isfile(image_file_names) and image_file_names.endswith('.json'):
            list_file = image_file_names
            with open(list_file) as f:
                image_file_names = json.load(f)
            print('Loaded {} image filenames from list file {}'.format(len(image_file_names),list_file))
            
        # A single image file
        elif os.path.isfile(image_file_names) and path_utils.is_image_file(image_file_names):
            image_file_names = [image_file_names]
            print('Processing image {}'.format(image_file_names[0]))
            
        else:        
            raise ValueError('image_file_names is a string, but is not a directory, a json ' + \
                             'list (.json), or an image file')
    
    if results is None:
        results = []

    already_processed = set([i['file'] for i in results])

    model_file = try_download_known_detector(model_file)
        
    print('GPU available: {}'.format(is_gpu_available(model_file)))
    
    if n_cores > 1 and is_gpu_available(model_file):
        
        print('Warning: multiple cores requested, but a GPU is available; parallelization across ' + \
              'GPUs is not currently supported, defaulting to one GPU')
        n_cores = 1

    if n_cores > 1 and use_image_queue:
        
        print('Warning: multiple cores requested, but the image queue is enabled; parallelization ' + \
              'with the image queue is not currently supported, defaulting to one worker')
        n_cores = 1
        
    if use_image_queue:
        
        assert checkpoint_frequency < 0, \
            'Using an image queue is not currently supported when checkpointing is enabled'
        assert len(results) == 0, \
            'Using an image queue with results loaded from a checkpoint is not currently supported'
        assert n_cores <= 1
        results = run_detector_with_image_queue(image_file_names, model_file, 
                                                confidence_threshold, quiet, 
                                                image_size=image_size)
        
    elif n_cores <= 1:

        # Load the detector
        start_time = time.time()
        detector = load_detector(model_file)
        elapsed = time.time() - start_time
        print('Loaded model in {}'.format(humanfriendly.format_timespan(elapsed)))

        # This is only used for console reporting, so it's OK that it doesn't
        # include images we might have loaded from a previous checkpoint
        count = 0

        for im_file in tqdm(image_file_names):

            # Will not add additional entries not in the starter checkpoint
            if im_file in already_processed:
                if not quiet:
                    print('Bypassing image {}'.format(im_file))
                continue

            count += 1

            result = process_image(im_file, detector, 
                                   confidence_threshold, quiet=quiet, 
                                   image_size=image_size, include_image_size=include_image_size,
                                   include_image_timestamp=include_image_timestamp,
                                   include_exif_data=include_exif_data)
            results.append(result)

            # Write a checkpoint if necessary
            if checkpoint_frequency != -1 and count % checkpoint_frequency == 0:
                
                print('Writing a new checkpoint after having processed {} images since '
                      'last restart'.format(count))
                
                write_checkpoint(checkpoint_path, results)
            
    else:
        
        # Multiprocessing is enabled at this point
        
        # When using multiprocessing, tell the workers to load the model on each
        # process, by passing the model_file string as the "model" argument to
        # process_images.
        detector = model_file

        print('Creating pool with {} cores'.format(n_cores))

        if len(already_processed) > 0:
            n_images_all = len(image_file_names)
            image_file_names = [fn for fn in image_file_names if fn not in already_processed]
            print('Loaded {} of {} images from checkpoint'.format(
                len(already_processed),n_images_all))
        
        # Divide images into chunks; we'll send one chunk to each worker process   
        image_batches = list(chunks_by_number_of_chunks(image_file_names, n_cores))
                
        pool = workerpool(n_cores)

        if checkpoint_path is not None:
            
            # Multiprocessing and checkpointing are both enabled at this point
            
            checkpoint_queue = Manager().Queue()
            
            # Pass the "results" array (which may already contain images loaded from an existing
            # checkpoint) to the checkpoint queue handler function, which will append results to 
            # the list as they become available.
            checkpoint_thread = Thread(target=checkpoint_queue_handler, 
                                       args=(checkpoint_path, checkpoint_frequency,
                                             checkpoint_queue, results), daemon=True)
            checkpoint_thread.start()

            pool.map(partial(process_images, detector=detector,
                                    confidence_threshold=confidence_threshold,
                                    image_size=image_size, 
                                    include_image_size=include_image_size,
                                    include_image_timestamp=include_image_timestamp,
                                    include_exif_data=include_exif_data,
                                    checkpoint_queue=checkpoint_queue), 
                                    image_batches)

            checkpoint_queue.put(None)

        else:
            
            # Multprocessing is enabled, but checkpointing is not
            
            new_results = pool.map(partial(process_images, detector=detector,
                                   confidence_threshold=confidence_threshold,image_size=image_size,
                                   include_image_size=include_image_size,
                                   include_image_timestamp=include_image_timestamp,
                                   include_exif_data=include_exif_data), 
                                   image_batches)

            new_results = list(itertools.chain.from_iterable(new_results))
            
            # Append the results we just computed to "results", which is *usually* empty, but will
            # be non-empty if we resumed from a checkpoint
            results += new_results

        # ...if checkpointing is/isn't enabled
        
    # ...if we're running (1) with image queue, (2) on one core, (3) on multiple cores
    
    # 'results' may have been modified in place, but we also return it for
    # backwards-compatibility.
    return results

# ...def load_and_run_detector_batch(...)


def checkpoint_queue_handler(checkpoint_path, checkpoint_frequency, checkpoint_queue, results):
    """
    Thread function to accumulate results and write checkpoints when checkpointing and
    multiprocessing are both enabled.
    """
    
    result_count = 0
    while True:
        result = checkpoint_queue.get()        
        if result is None:            
            break  
        
        result_count +=1
        results.append(result)

        if (checkpoint_frequency != -1) and (result_count % checkpoint_frequency == 0):
                
            print('Writing a new checkpoint after having processed {} images since '
                    'last restart'.format(result_count))
            
            write_checkpoint(checkpoint_path, results)


def write_checkpoint(checkpoint_path, results):
    """
    Writes the 'images' field in the dict 'results' to a json checkpoint file.
    """
    
    assert checkpoint_path is not None              
            
    # Back up any previous checkpoints, to protect against crashes while we're writing
    # the checkpoint file.
    checkpoint_tmp_path = None
    if os.path.isfile(checkpoint_path):
        checkpoint_tmp_path = checkpoint_path + '_tmp'
        shutil.copyfile(checkpoint_path,checkpoint_tmp_path)
        
    # Write the new checkpoint
    with open(checkpoint_path, 'w') as f:
        json.dump({'images': results}, f, indent=1, default=str)
        
    # Remove the backup checkpoint if it exists
    if checkpoint_tmp_path is not None:
        os.remove(checkpoint_tmp_path)


def get_image_datetime(image):
    """
    Returns the EXIF datetime from [image] (a PIL Image object), if available, as a string.
    
    [im_file] is used only for error reporting.
    """
    
    exif_tags = read_exif.read_pil_exif(image,exif_options)
    
    try:
        datetime_str = exif_tags['DateTimeOriginal']
        _ = time.strptime(datetime_str, '%Y:%m:%d %H:%M:%S')
        return datetime_str

    except Exception:
        return None        


def write_results_to_file(results, output_file, relative_path_base=None, 
                          detector_file=None, info=None, include_max_conf=False,
                          custom_metadata=None):
    """
    Writes list of detection results to JSON output file. Format matches:

    https://github.com/agentmorris/MegaDetector/tree/master/api/batch_processing#batch-processing-api-output-format

    Args
    - results: list of dict, each dict represents detections on one image
    - output_file: str, path to JSON output file, should end in '.json'
    - relative_path_base: str, path to a directory as the base for relative paths
    - detector_file: filename of the detector used to generate these results, only
        used to pull out a version number for the "info" field
    - info: dictionary to use instead of the default "info" field
    - include_max_conf: old files (version 1.2 and earlier) included a "max_conf" field
        in each image; this was removed in version 1.3.  Set this flag to force the inclusion
        of this field.
    - custom_metadata: additional data to include as info['custom_metadata'].  Typically
        a dictionary, but no format checks are performed.
        
    Returns the complete output dictionary that was written to the output file.
    """
    
    if relative_path_base is not None:
        results_relative = []
        for r in results:
            r_relative = copy.copy(r)
            r_relative['file'] = os.path.relpath(r_relative['file'], start=relative_path_base)
            results_relative.append(r_relative)
        results = results_relative

    # The typical case: we need to build the 'info' struct
    if info is None:
        
        info = { 
            'detection_completion_time': datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S'),
            'format_version': '1.3' 
        }
        
        if detector_file is not None:
            detector_filename = os.path.basename(detector_file)
            detector_version = get_detector_version_from_filename(detector_filename)
            detector_metadata = get_detector_metadata_from_version_string(detector_version)
            info['detector'] = detector_filename  
            info['detector_metadata'] = detector_metadata
        else:
            info['detector'] = 'unknown'
            info['detector_metadata'] = get_detector_metadata_from_version_string('unknown')
        
    # If the caller supplied the entire "info" struct
    else:
        
        if detector_file is not None:
            
            print('Warning (write_results_to_file): info struct and detector file ' + \
                  'supplied, ignoring detector file')

    if custom_metadata is not None:
        info['custom_metadata'] = custom_metadata
        
    # The 'max_detection_conf' field used to be included by default, and it caused all kinds
    # of headaches, so it's no longer included unless the user explicitly requests it.
    if not include_max_conf:
        for im in results:
            if 'max_detection_conf' in im:
                del im['max_detection_conf']
            
    final_output = {
        'images': results,
        'detection_categories': run_detector.DEFAULT_DETECTOR_LABEL_MAP,
        'info': info
    }
    
    with open(output_file, 'w') as f:
        json.dump(final_output, f, indent=1, default=str)
    print('Output file saved at {}'.format(output_file))
    
    return final_output

# ...def write_results_to_file(...)


#%% Interactive driver

if False:
    
    pass

    #%%
    
    checkpoint_path = None
    model_file = r'G:\temp\models\md_v4.1.0.pb'
    confidence_threshold = 0.1
    checkpoint_frequency = -1
    results = None
    ncores = 1
    use_image_queue = False
    quiet = False
    image_dir = r'G:\temp\demo_images\ssmini'
    image_size = None
    image_file_names = path_utils.find_images(image_dir, recursive=False)    
    
    start_time = time.time()
    
    results = load_and_run_detector_batch(model_file=model_file,
                                          image_file_names=image_file_names,
                                          checkpoint_path=checkpoint_path,
                                          confidence_threshold=confidence_threshold,
                                          checkpoint_frequency=checkpoint_frequency,
                                          results=results,
                                          n_cores=ncores,
                                          use_image_queue=use_image_queue,
                                          quiet=quiet,
                                          image_size=image_size)
    
    elapsed = time.time() - start_time
    
    print('Finished inference in {}'.format(humanfriendly.format_timespan(elapsed)))

    
#%% Command-line driver

def main():
    
    parser = argparse.ArgumentParser(
        description='Module to run a TF/PT animal detection model on lots of images')
    parser.add_argument(
        'detector_file',
        help='Path to detector model file (.pb or .pt).  Can also be the strings "MDV4", "MDV5A", or "MDV5B" to request automatic download.')
    parser.add_argument(
        'image_file',
        help='Path to a single image file, a JSON file containing a list of paths to images, or a directory')
    parser.add_argument(
        'output_file',
        help='Path to output JSON results file, should end with a .json extension')
    parser.add_argument(
        '--recursive',
        action='store_true',
        help='Recurse into directories, only meaningful if image_file points to a directory')
    parser.add_argument(
        '--output_relative_filenames',
        action='store_true',
        help='Output relative file names, only meaningful if image_file points to a directory')
    parser.add_argument(
        '--include_max_conf',
        action='store_true',
        help='Include the "max_detection_conf" field in the output')
    parser.add_argument(
        '--quiet',
        action='store_true',
        help='Suppress per-image console output')
    parser.add_argument(
        '--image_size',
        type=int,
        default=None,
        help=('Force image resizing to a (square) integer size (not recommended to change this)'))    
    parser.add_argument(
        '--use_image_queue',
        action='store_true',
        help='Pre-load images, may help keep your GPU busy; does not currently support ' + \
             'checkpointing.  Useful if you have a very fast GPU and a very slow disk.')
    parser.add_argument(
        '--threshold',
        type=float,
        default=run_detector.DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD,
        help="Confidence threshold between 0 and 1.0, don't include boxes below this " + \
            "confidence in the output file. Default is {}".format(
                run_detector.DEFAULT_OUTPUT_CONFIDENCE_THRESHOLD))
    parser.add_argument(
        '--checkpoint_frequency',
        type=int,
        default=-1,
        help='Write results to a temporary file every N images; default is -1, which ' + \
             'disables this feature')
    parser.add_argument(
        '--checkpoint_path',
        type=str,
        default=None,
        help='File name to which checkpoints will be written if checkpoint_frequency is > 0')    
    parser.add_argument(
        '--resume_from_checkpoint',
        type=str,
        default=None,
        help='Path to a JSON checkpoint file to resume from')
    parser.add_argument(
        '--allow_checkpoint_overwrite',
        action='store_true',
        help='By default, this script will bail if the specified checkpoint file ' + \
              'already exists; this option allows it to overwrite existing checkpoints')
    parser.add_argument(
        '--ncores',
        type=int,
        default=0,
        help='Number of cores to use; only applies to CPU-based inference')
    parser.add_argument(
        '--class_mapping_filename',
        type=str,
        default=None,
        help='Use a non-default class mapping, supplied in a .json file with a dictionary mapping' + \
            'int-strings to strings.  This will also disable the addition of "1" to all category ' + \
            'IDs, so your class mapping should start at zero.')
    parser.add_argument(
        '--include_image_size',
        action='store_true',
        help='Include image dimensions in output file'
    )
    parser.add_argument(
        '--include_image_timestamp',
        action='store_true',
        help='Include image datetime (if available) in output file'
    )
    parser.add_argument(
        '--include_exif_data',
        action='store_true',
        help='Include available EXIF data in output file'
    )
    parser.add_argument(
        '--overwrite_handling',
        type=str,
        default='overwrite',
        help='What should we do if the output file exists?  overwrite/skip/error (default overwrite)'
    )
    
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    # If the specified detector file is really the name of a known model, find 
    # (and possibly download) that model
    args.detector_file = try_download_known_detector(args.detector_file)
    
    assert os.path.exists(args.detector_file), \
        'detector file {} does not exist'.format(args.detector_file)
    assert 0.0 < args.threshold <= 1.0, 'Confidence threshold needs to be between 0 and 1'
    assert args.output_file.endswith('.json'), 'output_file specified needs to end with .json'
    if args.checkpoint_frequency != -1:
        assert args.checkpoint_frequency > 0, 'Checkpoint_frequency needs to be > 0 or == -1'
    if args.output_relative_filenames:
        assert os.path.isdir(args.image_file), \
            f'Could not find folder {args.image_file}, must supply a folder when ' + \
                '--output_relative_filenames is set'

    if os.path.exists(args.output_file):
        if args.overwrite_handling == 'overwrite':
            print('Warning: output file {} already exists and will be overwritten'.format(
                args.output_file))
        elif args.overwrite_handling == 'skip':
            print('Output file {} exists, returning'.format(
                args.output_file))
            return
        elif args.overwrite_handling == 'error':
            raise Exception('Output file {} exists'.format(args.output_file))
        else:
            raise ValueError('Illegal overwrite handling string {}'.format(args.overwrite_handling))

    # This is an experimental hack to allow the use of non-MD YOLOv5 models through
    # the same infrastructure; it disables the code that enforces MDv5-like class lists.
    if args.class_mapping_filename is not None:
        run_detector.USE_MODEL_NATIVE_CLASSES = True
        with open(args.class_mapping_filename,'r') as f:
            class_mapping = json.load(f)
        print('Loaded custom class mapping:')
        print(class_mapping)
        run_detector.DEFAULT_DETECTOR_LABEL_MAP = class_mapping
        
    # Load the checkpoint if available
    #
    # Relative file names are only output at the end; all file paths in the checkpoint are
    # still full paths.
    if args.resume_from_checkpoint is not None:
        assert os.path.exists(args.resume_from_checkpoint), \
            'File at resume_from_checkpoint specified does not exist'
        with open(args.resume_from_checkpoint) as f:
            print('Loading previous results from checkpoint file {}'.format(
                args.resume_from_checkpoint))
            saved = json.load(f)
        assert 'images' in saved, \
            'The checkpoint file does not have the correct fields; cannot be restored'
        results = saved['images']
        print('Restored {} entries from the checkpoint'.format(len(results)))
    else:
        results = []

    # Find the images to score; images can be a directory, may need to recurse
    if os.path.isdir(args.image_file):
        image_file_names = path_utils.find_images(args.image_file, args.recursive)
        if len(image_file_names) > 0:
            print('{} image files found in the input directory'.format(len(image_file_names)))                        
        else:
            if (args.recursive):
                print('No image files found in directory {}, exiting'.format(args.image_file))
            else:
                print('No image files found in directory {}, did you mean to specify '
                      '--recursive?'.format(
                    args.image_file))
            return
        
    # A json list of image paths
    elif os.path.isfile(args.image_file) and args.image_file.endswith('.json'):
        with open(args.image_file) as f:
            image_file_names = json.load(f)
        print('Loaded {} image filenames from list file {}'.format(
            len(image_file_names),args.image_file))
        
    # A single image file
    elif os.path.isfile(args.image_file) and path_utils.is_image_file(args.image_file):
        image_file_names = [args.image_file]
        print('Processing image {}'.format(args.image_file))
        
    else:        
        raise ValueError('image_file specified is not a directory, a json list, or an image file, '
                         '(or does not have recognizable extensions).')

    assert len(image_file_names) > 0, 'Specified image_file does not point to valid image files'
    assert os.path.exists(image_file_names[0]), \
        'The first image to be scored does not exist at {}'.format(image_file_names[0])

    output_dir = os.path.dirname(args.output_file)

    if len(output_dir) > 0:
        os.makedirs(output_dir,exist_ok=True)
        
    assert not os.path.isdir(args.output_file), 'Specified output file is a directory'

    # Test that we can write to the output_file's dir if checkpointing requested
    if args.checkpoint_frequency != -1:
        
        if args.checkpoint_path is not None:
            checkpoint_path = args.checkpoint_path
        else:
            checkpoint_path = os.path.join(output_dir,
                                           'checkpoint_{}.json'.format(
                                               datetime.utcnow().strftime("%Y%m%d%H%M%S")))
        
        # Don't overwrite existing checkpoint files, this is a sure-fire way to eventually
        # erase someone's checkpoint.
        if (checkpoint_path is not None) and (not args.allow_checkpoint_overwrite) \
            and (args.resume_from_checkpoint is None):
            
            assert not os.path.isfile(checkpoint_path), \
                f'Checkpoint path {checkpoint_path} already exists, delete or move it before ' + \
                're-using the same checkpoint path, or specify --allow_checkpoint_overwrite'

        
        # Confirm that we can write to the checkpoint path; this avoids issues where
        # we crash after several thousand images.
        #
        # But actually, commenting this out for now... the scenario where we are resuming from a 
        # checkpoint, then immediately overwrite that checkpoint with empty data is higher-risk
        # than the annoyance of crashing a few minutes after starting a job.
        if False:
            with open(checkpoint_path, 'w') as f:
                json.dump({'images': []}, f)
                
        print('The checkpoint file will be written to {}'.format(checkpoint_path))
        
    else:
        
        checkpoint_path = None

    start_time = time.time()

    results = load_and_run_detector_batch(model_file=args.detector_file,
                                          image_file_names=image_file_names,
                                          checkpoint_path=checkpoint_path,
                                          confidence_threshold=args.threshold,
                                          checkpoint_frequency=args.checkpoint_frequency,
                                          results=results,
                                          n_cores=args.ncores,
                                          use_image_queue=args.use_image_queue,
                                          quiet=args.quiet,
                                          image_size=args.image_size,
                                          class_mapping_filename=args.class_mapping_filename,
                                          include_image_size=args.include_image_size,
                                          include_image_timestamp=args.include_image_timestamp,
                                          include_exif_data=args.include_exif_data)

    elapsed = time.time() - start_time
    images_per_second = len(results) / elapsed
    print('Finished inference for {} images in {} ({:.2f} images per second)'.format(
        len(results),humanfriendly.format_timespan(elapsed),images_per_second))

    relative_path_base = None
    if args.output_relative_filenames:
        relative_path_base = args.image_file
    write_results_to_file(results, args.output_file, relative_path_base=relative_path_base,
                          detector_file=args.detector_file,include_max_conf=args.include_max_conf)

    if checkpoint_path and os.path.isfile(checkpoint_path):
        os.remove(checkpoint_path)
        print('Deleted checkpoint file {}'.format(checkpoint_path))

    print('Done!')


if __name__ == '__main__':
    main()



########
#
# ct_utils.py
#
# Utility functions that don't depend on other things in this repo.
#
########

#%% Imports and constants

import argparse
import inspect
import json
import math
import os

import jsonpickle
import numpy as np

# List of file extensions we'll consider images; comparisons will be case-insensitive
# (i.e., no need to include both .jpg and .JPG on this list).
image_extensions = ['.jpg', '.jpeg', '.gif', '.png']


#%% Functions

def truncate_float_array(xs, precision=3):
    """
    Vectorized version of truncate_float(...)

    Args:
    xs        (list of float) List of floats to truncate
    precision (int)           The number of significant digits to preserve, should be
                              greater or equal 1
    """

    return [truncate_float(x, precision=precision) for x in xs]


def truncate_float(x, precision=3):
    """
    Truncates a floating-point value to a specific number of significant digits.
    
    For example: truncate_float(0.0003214884) --> 0.000321
    
    This function is primarily used to achieve a certain float representation
    before exporting to JSON.

    Args:
    x         (float) Scalar to truncate
    precision (int)   The number of significant digits to preserve, should be
                      greater or equal 1
    """

    assert precision > 0

    if np.isclose(x, 0):
        
        return 0
    
    else:
        
        # Determine the factor, which shifts the decimal point of x
        # just behind the last significant digit.
        factor = math.pow(10, precision - 1 - math.floor(math.log10(abs(x))))
        
        # Shift decimal point by multiplicatipon with factor, flooring, and
        # division by factor.
        return math.floor(x * factor)/factor


def args_to_object(args: argparse.Namespace, obj: object) -> None:
    """
    Copies all fields from a Namespace (i.e., the output from parse_args) to an
    object. Skips fields starting with _. Does not check existence in the target
    object.

    Args:
        args: argparse.Namespace
        obj: class or object whose whose attributes will be updated
    """
    
    for n, v in inspect.getmembers(args):
        if not n.startswith('_'):
            setattr(obj, n, v)


def pretty_print_object(obj, b_print=True):
    """
    Prints an arbitrary object as .json
    """

    # _ = pretty_print_object(obj)

    # Sloppy that I'm making a module-wide change here...
    jsonpickle.set_encoder_options('json', sort_keys=True, indent=2)
    a = jsonpickle.encode(obj)
    s = '{}'.format(a)
    if b_print:
        print(s)
    return s


def is_list_sorted(L,reverse=False):
    """
    Returns true if the list L appears to be sorted, otherwise False.
    
    Calling is_list_sorted(L,reverse=True) is the same as calling
    is_list_sorted(L.reverse(),reverse=False).
    """
    
    if reverse:
        return all(L[i] >= L[i + 1] for i in range(len(L)-1))
    else:
        return all(L[i] <= L[i + 1] for i in range(len(L)-1))
        

def write_json(path, content, indent=1):
    """
    Standardized wrapper for json.dump
    """
    
    with open(path, 'w') as f:
        json.dump(content, f, indent=indent)


def is_image_file(s):
    """
    Checks a file's extension against a hard-coded set of image file extensions; 
    return True if it appears to be an image.
    """

    ext = os.path.splitext(s)[1]
    return ext.lower() in image_extensions


def convert_yolo_to_xywh(yolo_box):
    """
    Converts a YOLO format bounding box to [x_min, y_min, width_of_box, height_of_box].

    Args:
        yolo_box: bounding box of format [x_center, y_center, width_of_box, height_of_box].

    Returns:
        bbox with coordinates represented as [x_min, y_min, width_of_box, height_of_box].
    """
    
    x_center, y_center, width_of_box, height_of_box = yolo_box
    x_min = x_center - width_of_box / 2.0
    y_min = y_center - height_of_box / 2.0
    return [x_min, y_min, width_of_box, height_of_box]


def convert_xywh_to_tf(api_box):
    """
    Converts an xywh bounding box to an [y_min, x_min, y_max, x_max] box that the TensorFlow
    Object Detection API uses

    Args:
        api_box: bbox output by the batch processing API [x_min, y_min, width_of_box, height_of_box]

    Returns:
        bbox with coordinates represented as [y_min, x_min, y_max, x_max]
    """
    
    x_min, y_min, width_of_box, height_of_box = api_box
    x_max = x_min + width_of_box
    y_max = y_min + height_of_box
    return [y_min, x_min, y_max, x_max]


def convert_xywh_to_xyxy(api_bbox):
    """
    Converts an xywh bounding box to an xyxy bounding box.

    Note that this is also different from the TensorFlow Object Detection API coords format.
    Args:
        api_bbox: bbox output by the batch processing API [x_min, y_min, width_of_box, height_of_box]

    Returns:
        bbox with coordinates represented as [x_min, y_min, x_max, y_max]
    """

    x_min, y_min, width_of_box, height_of_box = api_bbox
    x_max, y_max = x_min + width_of_box, y_min + height_of_box
    return [x_min, y_min, x_max, y_max]


def get_iou(bb1, bb2):
    """
    Calculates the Intersection over Union (IoU) of two bounding boxes.

    Adapted from:
        
    https://stackoverflow.com/questions/25349178/calculating-percentage-of-bounding-box-overlap-for-image-detector-evaluation

    Args:
        bb1: [x_min, y_min, width_of_box, height_of_box]
        bb2: [x_min, y_min, width_of_box, height_of_box]

    Returns:
        intersection_over_union, a float in [0, 1]
    """

    bb1 = convert_xywh_to_xyxy(bb1)
    bb2 = convert_xywh_to_xyxy(bb2)

    assert bb1[0] < bb1[2], 'Malformed bounding box (x2 >= x1)'
    assert bb1[1] < bb1[3], 'Malformed bounding box (y2 >= y1)'

    assert bb2[0] < bb2[2], 'Malformed bounding box (x2 >= x1)'
    assert bb2[1] < bb2[3], 'Malformed bounding box (y2 >= y1)'

    # Determine the coordinates of the intersection rectangle
    x_left = max(bb1[0], bb2[0])
    y_top = max(bb1[1], bb2[1])
    x_right = min(bb1[2], bb2[2])
    y_bottom = min(bb1[3], bb2[3])

    if x_right < x_left or y_bottom < y_top:
        return 0.0

    # The intersection of two axis-aligned bounding boxes is always an
    # axis-aligned bounding box
    intersection_area = (x_right - x_left) * (y_bottom - y_top)

    # Compute the area of both AABBs
    bb1_area = (bb1[2] - bb1[0]) * (bb1[3] - bb1[1])
    bb2_area = (bb2[2] - bb2[0]) * (bb2[3] - bb2[1])

    # Compute the intersection over union by taking the intersection
    # area and dividing it by the sum of prediction + ground-truth
    # areas - the intersection area.
    iou = intersection_area / float(bb1_area + bb2_area - intersection_area)
    assert iou >= 0.0, 'Illegal IOU < 0'
    assert iou <= 1.0, 'Illegal IOU > 1'
    return iou


def _get_max_conf_from_detections(detections):
    """
    Internal function used by get_max_conf(); don't call this directly.
    """
    
    max_conf = 0.0
    if detections is not None and len(detections) > 0:
        confidences = [det['conf'] for det in detections]
        max_conf = max(confidences)
    return max_conf


def get_max_conf(im):
    """
    Given an image dict in the format used by the batch API, compute the maximum detection
    confidence for any class.  Returns 0.0 (not None) if there was a failure and 'detections'
    isn't present.
    """
    
    max_conf = 0.0
    if 'detections' in im and im['detections'] is not None and len(im['detections']) > 0:
        max_conf = _get_max_conf_from_detections(im['detections'])
    return max_conf
